from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.parallel.abstract import Extract2DParallel
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DParallelEPER(Extract2DParallel):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D parallel EPER regions from the `region_list` containing signal (e.g. the charge
        injection regions of charge injection data), between two input `pixels` indexes.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the parallel EPERs
        (e.g. the FPRs) are extracted.

        The method includes a unique behaviour overide, where if `pixels_from_end=-1` is input the method will return
        the region lists containg the full parallel EPER regions after each charge injection block (accounting for
        the possibility that charge injection regions are not spaced uniformly).

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

              [ppppppppppppppppppppp]
              [ppppppppppppppppppppp]
         [...][t1t1t1t1t1t1t1t1t1t1t][sss]
         [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1c][sss]     |
        | [...][t0t0t0t0t0t0t0t0t0t0t][sss]    | Direction
        Par [...][0t0t0t0t0t0t0t0t0t0t0][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/ [...][cc0ccc0cccc0cccc0cccc][sss]   \/

        []     [=====================]
              <---------Ser----------

        The extracted regions correspond to the first parallel EPER all charge injection regions:

        region_list[0] = [2, 4, 3, 21] (serial prescan is 3 pixels)
        region_list[1] = [6, 7, 3, 21] (serial prescan is 3 pixels)

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        parallel EPER of each charge injection region:

        array_2d_list[0] = [t0t0t0tt0t0t0t0t0t0t0]
        array_2d_list[1] = [1t1t1t1t1t1t1t1t1t1t1]

        Parameters
        ----------
        settings
           The settings used to extract the parallel EPERs, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """

        pixels = settings.pixels

        region_list = []

        if settings.pixels_from_end is not None:
            parallel_row_spaces = self.parallel_rows_between_regions + [
                self.parallel_rows_to_array_edge
            ]

        for i, region in enumerate(self.region_list):
            if settings.pixels_from_end is not None:
                if settings.pixels_from_end == -1:
                    pixels = (
                        0,
                        parallel_row_spaces[i],
                    )

                else:
                    pixels = (
                        parallel_row_spaces[i] - settings.pixels_from_end,
                        parallel_row_spaces[i],
                    )

            region_list.append(region.parallel_trailing_region_from(pixels=pixels))

        return region_list

    def array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
        Extract from an `Array2D` the parallel EPERs and return them as a new `Array2D` where everything else (e.g.
        the charge injection regions, serial EPERs) are zeros.

        The dimensions of the input array therefore do not change (unlike other ``Extract`` methods).

        Negative pixel values can be input into the `pixels` tuple, whereby rows in front of the parallel EPERs (e.g.
        the FPR) are extracted.

        The diagram below illustrates the extraction:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

             [tptpptptptpptpptpptpt]
             [tptptptpptpttptptptpt]
        [...][ttttttttttttttttttttt][sss]
        [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        Par[...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
        |/ [...][ccccccccccccccccccccc][sss]   \/

        []     [=====================]
             <--------Ser---------

        The extracted array keeps only the parallel EPERs, everything else become 0s:

             [tptpptptptpptpptpptpt]
             [tptptptpptpttptptptpt]
          [000][ttttttttttttttttttttt][000]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][ttttttttttttttttttttt][000]    | Direction
        Par[000][ttttttttttttttttttttt][000]    | of
        | [000][000000000000000000000][000]    | clocking
        |/ [000][000000000000000000000][000]   \/

        []     [=====================]
             <--------Ser---------
        """

        array_2d = array.native.copy()

        for region in self.region_list:
            array_2d[region.slice] = 0.0

        array_2d.native[self.serial_prescan.slice] = 0.0
        array_2d.native[self.serial_overscan.slice] = 0.0

        return array_2d

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        """
        The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
        CTI modeling in 1D.

        This is performed by binning up the data via the `binned_array_1d_from` function.

        In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
        charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

        The function returns the this region if the 1D dataset is extracted from the parallel EPERs. The
        charge region is only included if there are negative entries in the `pixels` tuple, meaning that pixels
        before the EPERs (e.g. the FPR) are extracted.

        Parameters
        ----------
        settings
           The settings used to extract the parallel EPERs, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """
        return extract_2d_util.binned_region_1d_eper_from(pixels=settings.pixels)



from autofit.plot.samples_plotters import SamplesPlotter
from autofit.non_linear.search.nest.dynesty.plotter import DynestyPlotter
from autofit.non_linear.search.nest.nautilus.plotter import NautilusPlotter
from autofit.non_linear.search.nest.ultranest.plotter import UltraNestPlotter
from autofit.non_linear.search.mcmc.emcee.plotter import EmceePlotter
from autofit.non_linear.search.mcmc.zeus.plotter import ZeusPlotter
from autofit.non_linear.search.optimize.pyswarms.plotter import PySwarmsPlotter

from autoarray.plot.wrap.base import Axis
from autoarray.plot.wrap.base import Units
from autoarray.plot.wrap.base import Figure
from autoarray.plot.wrap.base import Cmap
from autoarray.plot.wrap.base import Colorbar
from autoarray.plot.wrap.base import ColorbarTickParams
from autoarray.plot.wrap.base import TickParams
from autoarray.plot.wrap.base import YTicks
from autoarray.plot.wrap.base import XTicks
from autoarray.plot.wrap.base import Title
from autoarray.plot.wrap.base import YLabel
from autoarray.plot.wrap.base import XLabel
from autoarray.plot.wrap.base import Legend
from autoarray.plot.wrap.base import Output

from autoarray.plot.wrap.one_d import YXPlot
from autoarray.plot.wrap.two_d import ArrayOverlay
from autoarray.plot.wrap.two_d import GridScatter
from autoarray.plot.wrap.two_d import GridPlot
from autoarray.plot.wrap.two_d import VectorYXQuiver
from autoarray.plot.wrap.two_d import PatchOverlay
from autoarray.plot.wrap.two_d import VoronoiDrawer
from autoarray.plot.wrap.two_d import OriginScatter
from autoarray.plot.wrap.two_d import MaskScatter
from autoarray.plot.wrap.two_d import BorderScatter
from autoarray.plot.wrap.two_d import PositionsScatter
from autoarray.plot.wrap.two_d import IndexScatter
from autoarray.plot.wrap.two_d import MeshGridScatter
from autoarray.plot.wrap.two_d import ParallelOverscanPlot
from autoarray.plot.wrap.two_d import SerialPrescanPlot
from autoarray.plot.wrap.two_d import SerialOverscanPlot

from autoarray.plot.mat_plot.one_d import MatPlot1D
from autoarray.plot.include.one_d import Include1D
from autoarray.plot.visuals.one_d import Visuals1D
from autoarray.plot.mat_plot.two_d import MatPlot2D
from autoarray.plot.include.two_d import Include2D
from autoarray.plot.visuals.two_d import Visuals2D

from autoarray.structures.plot.structure_plotters import YX1DPlotter
from autoarray.structures.plot.structure_plotters import YX1DPlotter as Array1DPlotter
from autoarray.structures.plot.structure_plotters import Array2DPlotter

from autoarray.plot.multi_plotters import MultiFigurePlotter
from autoarray.plot.multi_plotters import MultiYX1DPlotter

from autocti.dataset_1d.plot.dataset_1d_plotters import Dataset1DPlotter
from autocti.dataset_1d.plot.fit_plotters import FitDataset1DPlotter
from autocti.charge_injection.plot.imaging_ci_plotters import ImagingCIPlotter
from autocti.charge_injection.plot.fit_ci_plotters import FitImagingCIPlotter



import autoarray as aa

from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D


class FitDataset1D(aa.FitDataset):
    def __init__(self, dataset: Dataset1D, post_cti_data):
        """
        Fit a 1D CTI dataset with model cti data.

        Parameters
        ----------
        dataset
            The charge injection image that is fitted.
        post_cti_data
            The `pre_cti_data` with cti added to it via the clocker and a CTI model.
        """

        super().__init__(dataset=dataset, use_mask_in_fit=True)

        self.post_cti_data = post_cti_data

    @property
    def dataset_1d(self) -> Dataset1D:
        return self.dataset

    @property
    def mask(self) -> aa.Mask1D:
        return self.dataset.mask

    @property
    def model_data(self) -> aa.Array1D:
        return self.post_cti_data

    @property
    def pre_cti_data(self) -> aa.Array1D:
        return self.dataset.pre_cti_data



from __future__ import annotations
import numpy as np
import math
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

if TYPE_CHECKING:
    from autocti.extract.two_d.master import Extract2DMaster

import autoarray as aa

from autoarray.layout import layout_util

from autocti.instruments.euclid.layout import Layout2DEuclid
from autocti.layout.two_d import Layout2D

from autocti.charge_injection import ci_util


class Layout2DCI(Layout2D):
    def __init__(
        self,
        shape_2d: Tuple[int, int],
        region_list: aa.type.Region2DList,
        original_roe_corner: Tuple[int, int] = (1, 0),
        parallel_overscan: Optional[aa.type.Region2DLike] = None,
        serial_prescan: Optional[aa.type.Region2DLike] = None,
        serial_overscan: Optional[aa.type.Region2DLike] = None,
        electronics: Optional["ElectronicsCI"] = None,
    ):
        """
        A charge injection layout, which defines the regions charge injections appear on a charge injection image.

        It also contains over regions of the image, for example the serial prescan, overscan and paralle overscan.

        Parameters
        ----------
        shape_2d
            The two dimensional shape of the charge injection imaging, corresponding to the number of rows (pixels
            in parallel direction) and columns (pixels in serial direction).
        region_list
            Integer pixel coordinates specifying the corners of each charge injection region (top-row, bottom-row,
            left-column, right-column).
        original_roe_corner
            The original read-out electronics corner of the charge injeciton imaging, which is internally rotated to a
            common orientation in **PyAutoCTI**.
        parallel_overscan
            Integer pixel coordinates specifying the corners of the parallel overscan (top-row, bottom-row,
            left-column, right-column).
        serial_prescan
            Integer pixel coordinates specifying the corners of the serial prescan (top-row, bottom-row,
            left-column, right-column).
        serial_overscan
            Integer pixel coordinates specifying the corners of the serial overscan (top-row, bottom-row,
            left-column, right-column).
        electronics
            The charge injection electronics parameters of the image (e.g. the IG1 and IG2 voltages).
        """

        super().__init__(
            shape_2d=shape_2d,
            region_list=region_list,
            original_roe_corner=original_roe_corner,
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

        self.electronics = electronics

    @property
    def extract(self) -> Extract2DMaster:
        from autocti.extract.two_d.master import Extract2DMaster

        return Extract2DMaster(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self.parallel_overscan,
            serial_prescan=self.serial_prescan,
            serial_overscan=self.serial_overscan,
        )

    @classmethod
    def from_euclid_fits_header(cls, ext_header):
        serial_overscan_size = ext_header.get("OVRSCANX", default=None)
        serial_prescan_size = ext_header.get("PRESCANX", default=None)
        serial_size = ext_header.get("NAXIS1", default=None)
        parallel_size = ext_header.get("NAXIS2", default=None)

        electronics = ElectronicsCI.from_ext_header(ext_header=ext_header)

        layout = Layout2DEuclid.from_fits_header(ext_header=ext_header)

        region_ci_list = ci_util.region_list_ci_via_electronics_from(
            injection_start=electronics.injection_start,
            injection_on=electronics.injection_on,
            injection_off=electronics.injection_off,
            injection_total=electronics.injection_total,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
            serial_size=serial_size,
            parallel_size=parallel_size,
            roe_corner=layout.original_roe_corner,
        )

        region_ci_list = [
            layout_util.rotate_region_via_roe_corner_from(
                region=region_ci,
                shape_native=layout.shape_2d,
                roe_corner=layout.original_roe_corner,
            )
            for region_ci in region_ci_list
        ]

        return cls(
            shape_2d=(parallel_size, serial_size),
            region_list=region_ci_list,
            original_roe_corner=layout.original_roe_corner,
            parallel_overscan=layout.parallel_overscan,
            serial_prescan=layout.serial_prescan,
            serial_overscan=layout.serial_overscan,
            electronics=electronics,
        )

    def pre_cti_data_uniform_from(
        self, norm: float, pixel_scales: aa.type.PixelScales
    ) -> aa.Array2D:
        """
        Use this charge injection layout to generate a pre-cti charge injection image.

        This is performed by going to its charge injection regions and adding an input charge injection
        normalization value.

        Parameters
        ----------
        norm
            The normalization of the charge injection region.
        pixel_scales
            The (y,x) scaled units to pixel units conversion factors of every pixel. If this is input as a `float`,
            it is converted to a (float, float) structure.
        """

        pre_cti_data = np.zeros(self.shape_2d)

        for region in self.region_list:
            pre_cti_data[region.slice] += norm

        return aa.Array2D.no_mask(values=pre_cti_data, pixel_scales=pixel_scales)

    def pre_cti_data_non_uniform_from(
        self,
        injection_norm_list: List[float],
        pixel_scales: aa.type.PixelScales,
        row_slope: Optional[float] = 0.0,
    ) -> aa.Array2D:
        """
        Use this charge injection layout to generate a pre-cti charge injection image. This is performed by going 
        to its charge injection regions and adding an input normalization value to each column, which are
        passed in as a list.

        For one column of a non-uniform charge injection pre-cti image, this function assumes that each non-uniform 
        charge injection region has the same overall normalization value. This assumes the spikes / troughs in the
        injection current that cause non-uniformity occur in an identical fashion for each charge injection region.

        Non-uniformity across the rows of a charge injection layout_ci is due to a drop-off in voltage in the current.
        Therefore, it appears smooth and be modeled as an analytic function, which this code assumes is a
        power-law with slope `row_slope`.

        Parameters
        ----------
        shape_native
            The image_shape of the pre_cti_datas to be created.
        row_slope
            The power-law slope of non-uniformity in the row charge injection profile.
        ci_seed
            Input ci_seed for the random number generator to give reproducible results. A new ci_seed is always used for each \
            pre_cti_datas, ensuring each non-uniform ci_region has the same column non-uniformity layout_ci.
        """

        pre_cti_data = np.zeros(self.shape_2d)

        for region in self.region_list:
            pre_cti_data[region.slice] += ci_util.region_ci_from(
                region_dimensions=region.shape,
                injection_norm_list=injection_norm_list,
                row_slope=row_slope,
            )

        return aa.Array2D.no_mask(values=pre_cti_data, pixel_scales=pixel_scales)

    def pre_cti_data_non_uniform_via_lists_from(
        self,
        injection_norm_lists: List[List[float]],
        pixel_scales: aa.type.PixelScales,
        row_slope: Optional[float] = 0.0,
    ) -> aa.Array2D:
        """
        Use this charge injection layout to generate a pre-cti charge injection image. This is performed by going
        to its charge injection regions and adding an input normalization value to each column, which are
        passed in as a list.

        For one column of a non-uniform charge injection pre-cti image, this function assumes that each non-uniform
        charge injection region can have a different normalization value. The normalizztion values `injection_norm_lists`
        are passed as a list of lists so that the normalization of each injection in each region can be specified.

        Non-uniformity across the rows of a charge injection layout_ci is due to a drop-off in voltage in the current.
        Therefore, it appears smooth and be modeled as an analytic function, which this code assumes is a
        power-law with slope `row_slope`.

        Parameters
        ----------
        shape_native
            The image_shape of the pre_cti_datas to be created.
        row_slope
            The power-law slope of non-uniformity in the row charge injection profile.
        ci_seed
            Input ci_seed for the random number generator to give reproducible results. A new ci_seed is always used for each \
            pre_cti_datas, ensuring each non-uniform ci_region has the same column non-uniformity layout_ci.
        """

        pre_cti_data = np.zeros(self.shape_2d)

        for region_index, region in enumerate(self.region_list):
            pre_cti_data[region.slice] += ci_util.region_ci_from(
                region_dimensions=region.shape,
                injection_norm_list=injection_norm_lists[region_index],
                row_slope=row_slope,
            )

        return aa.Array2D.no_mask(values=pre_cti_data, pixel_scales=pixel_scales)

    def noise_map_non_uniform_from(
        self,
        injection_std_list: List[float],
        pixel_scales: aa.type.PixelScales,
        read_noise: float = 0.0,
    ) -> aa.Array2D:
        """
        Use this charge injection layout to generate the noise-map of a charge injection image. This is performed by
        going to its charge injection regions and adding an input RMS standard deviation value to each column, which
        are passed in as a list.

        For one column of a non-uniform charge injection noise-map, this function assumes that each non-uniform
        charge injection region has the same overall noise value.

        A read-noise can also be be included when creating the noise map, which is added in quadrature to the
        charge injection noise values.
        """

        noise_map = np.full(fill_value=read_noise, shape=self.shape_2d)

        for region in self.region_list:
            noise_region = ci_util.region_ci_from(
                region_dimensions=region.shape,
                injection_norm_list=injection_std_list,
            )

            noise_map[region.slice] = np.sqrt(
                np.square(read_noise) + np.square(noise_region)
            )

        return aa.Array2D.no_mask(values=noise_map, pixel_scales=pixel_scales)

    def noise_map_non_uniform_via_lists_from(
        self,
        injection_std_lists: List[List[float]],
        pixel_scales: aa.type.PixelScales,
        read_noise: float = 0.0,
    ) -> aa.Array2D:
        """
        Use this charge injection layout to generate the noise-map of a charge injection image. This is performed by
        going to its charge injection regions and adding an input RMS standard deviation value to each column, which
        are passed in as a list.

        For one column of a non-uniform charge injection pre-cti image, this function assumes that each non-uniform
        charge injection region can have a different noise value. The alues `injection_std_lists` are passed as a
        list of lists so that the RMS standard deviation of each injection in each region can be specified.

        A read-noise can also be be included when creating the noise map, which is added in quadrature to the
        charge injection noise values.
        """

        noise_map = np.full(fill_value=read_noise, shape=self.shape_2d)

        for region_index, region in enumerate(self.region_list):
            noise_region = ci_util.region_ci_from(
                region_dimensions=region.shape,
                injection_norm_list=injection_std_lists[region_index],
            )

            noise_map[region.slice] = np.sqrt(
                np.square(read_noise) + np.square(noise_region)
            )

        return aa.Array2D.no_mask(values=noise_map, pixel_scales=pixel_scales)


class ElectronicsCI:
    def __init__(
        self,
        injection_on: Optional[int] = None,
        injection_off: Optional[int] = None,
        injection_start: Optional[int] = None,
        injection_end: Optional[int] = None,
        ig_1: Optional[float] = None,
        ig_2: Optional[float] = None,
    ):
        """
        Stores the electronics parameters contained for a charge injection line image, with this class currently
        specific to those in Euclid.

        These are extracted from the .fits file header of a charge injection image, with the original .fits
        headers included in the `as_ext_header_dict` property.

        The `injection_on` and `injection_off` parameters determine the number of pixels the charge injection is
        held on and then off for. The `v_start` and `v_end` parameters define the pixels where the charge injection
        starts and ends.

        For example, take a CCD which has 1820 rows of pixels, where:

        - `injection_on=100`
        - `injection_off=200`
        - `v_start=10`
        - `v_end`=1810

        Starting from row 10, for every 300 rows of pixels there first 100 pixels will contain charge injection and
        the remaining 200 rows will not (they will contain EPER trails). This pattern will be repeated 6 times over
        the next 1800 pixels of the CCD with the charge injection ending at 1810.

        NOTE: The charge injection electrons have the following four parameters:

        VSTART_CHJ_INJ
        VEND_CHJ_INJ
        VSTART
        VEND

        I do not yet know which of these maps to which fits header. I am currently assuming all 4 correspond to
        `v_start` and `v_end`, albeit their functionality is not used specifically.

        Parameters
        ----------
        injection_on
            The number of rows of pixels the charge injection is held on for per charge injection region.
        injection_off
            The number of rows of pixels the charge injection is held off for per charge injection region.
        injection_start
            The pixel row where the charge injection begins.
        injection_end
            The pixel row where the charge injection ends.
        ig_1
            The voltage of injection gate 1.
        ig_2
            The voltage of injection gate 2.
        """
        self.injection_on = injection_on
        self.injection_off = injection_off
        self.injection_start = injection_start
        self.injection_end = injection_end
        self.ig_1 = ig_1
        self.ig_2 = ig_2

    @classmethod
    def from_ext_header(cls, ext_header: Dict) -> "ElectronicsCI":
        """
        Creates the charge injection electronics from a Euclid charge injection imaging .fits header.

        Parameters
        ----------
        ext_header
            The .fits header dictionary of a Euclid charge injection image.
        """
        injection_on = ext_header["CI_IJON"]
        injection_off = ext_header["CI_IJOFF"]
        injection_start = ext_header["CI_VSTAR"]
        injection_end = ext_header["CI_VEND"]

        return ElectronicsCI(
            injection_on=injection_on,
            injection_off=injection_off,
            injection_start=injection_start,
            injection_end=injection_end,
        )

    @property
    def as_ext_header_dict(self) -> Dict:
        """
        Returns the charge injection electronics as a dictionary which is representative of the parameter values
        stored in a Euclid charge injection .fits image.
        """
        return {
            "CI_IJON": self.injection_on,
            "CI_IJOFF": self.injection_off,
            "CI_VSTAR": self.injection_start,
            "CI_VEND": self.injection_end,
        }

    @property
    def injection_total(self) -> int:
        """
        The total number of charge injection regions for these electronics settings.
        """

        injection_range = self.injection_end - self.injection_start

        for injection_total in range(100):
            total_pixels = math.floor(
                (injection_total + 1) * (self.injection_on)
                + injection_total * self.injection_off
            )

            if total_pixels > injection_range:
                return injection_total



from typing import Callable

from autoconf import conf

import autoarray.plot as aplt

from autoarray.plot.auto_labels import AutoLabels
from autoarray.dataset.plot.imaging_plotters import ImagingPlotterMeta

from autocti.plot.abstract_plotters import Plotter
from autocti.charge_injection.imaging.imaging import ImagingCI


class ImagingCIPlotter(Plotter):
    def __init__(
        self,
        dataset: ImagingCI,
        mat_plot_2d: aplt.MatPlot2D = aplt.MatPlot2D(),
        visuals_2d: aplt.Visuals2D = aplt.Visuals2D(),
        include_2d: aplt.Include2D = aplt.Include2D(),
        mat_plot_1d: aplt.MatPlot1D = aplt.MatPlot1D(),
        visuals_1d: aplt.Visuals1D = aplt.Visuals1D(),
        include_1d: aplt.Include1D = aplt.Include1D(),
        residuals_symmetric_cmap: bool = True,
    ):
        """
        Plots the attributes of `Imaging` objects using the matplotlib method `imshow()` and many other matplotlib
        functions which customize the plot's appearance.

        The `mat_plot_2d` attribute wraps matplotlib function calls to make the figure. By default, the settings
        passed to every matplotlib function called are those specified in the `config/visualize/mat_wrap/*.ini` files,
        but a user can manually input values into `MatPlot2d` to customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals2D` object. Attributes may be extracted from
        the `Imaging` and plotted via the visuals object, if the corresponding entry is `True` in the `Include2D`
        object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        dataset
            The charge injection imaging dataset the plotter plots.
        mat_plot_2d
            Contains objects which wrap the matplotlib function calls that make 2D plots.
        visuals_2d
            Contains 2D visuals that can be overlaid on 2D plots.
        include_2d
            Specifies which attributes of the `Imaging` are extracted and plotted as visuals for 2D plots.
        mat_plot_1d
            Contains objects which wrap the matplotlib function calls that make 1D plots.
        visuals_1d
            Contains 1D visuals that can be overlaid on 1D plots.
        include_1d
            Specifies which attributes of the `ImagingCI` are extracted and plotted as visuals for 1D plots.
        residuals_symmetric_cmap
            If true, the `pre_cti_residual_map` is plotted with a symmetric color map such that `abs(vmin) = abs(vmax)`.
        """
        super().__init__(
            dataset=dataset,
            mat_plot_2d=mat_plot_2d,
            include_2d=include_2d,
            visuals_2d=visuals_2d,
        )

        self.visuals_1d = visuals_1d
        self.include_1d = include_1d
        self.mat_plot_1d = mat_plot_1d

        self._imaging_meta_plotter = ImagingPlotterMeta(
            dataset=self.dataset,
            get_visuals_2d=self.get_visuals_2d,
            mat_plot_2d=self.mat_plot_2d,
            include_2d=self.include_2d,
            visuals_2d=self.visuals_2d,
        )

        self.residuals_symmetric_cmap = residuals_symmetric_cmap

    def get_visuals_1d(self):
        return self.visuals_1d

    def get_visuals_2d(self):
        return self.get_2d.via_mask_from(mask=self.dataset.mask)

    @property
    def extract_region_from(self) -> Callable:
        return self.dataset.layout.extract_region_from

    @property
    def extract_region_noise_map_from(self) -> Callable:
        return self.dataset.layout.extract_region_noise_map_from

    def figures_2d(
        self,
        data: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        pre_cti_data_residual_map: bool = False,
        cosmic_ray_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `ImagingCI` object in 2D.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        data
            Whether to make a 2D plot (via `imshow`) of the image data.
        noise_map
            Whether to make a 2D plot (via `imshow`) of the noise map.
        signal_to_noise_map
            Whether to make a 2D plot (via `imshow`) of the signal-to-noise map.
        pre_cti_data
            Whether to make a 2D plot (via `imshow`) of the pre-cti data.
        pre_cti_data_residual_map
            Whether to make a 2D plot (via `imshow`) of the pre-cti data residual-map.
        cosmic_ray_map
            Whether to make a 2D plot (via `imshow`) of the cosmic ray map.
        """

        title_str = self.title_str_2d_from()

        self._imaging_meta_plotter.figures_2d(
            data=data,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            title_str=title_str,
        )

        if pre_cti_data:
            self.mat_plot_2d.plot_array(
                array=self.dataset.pre_cti_data,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(
                    title=title_str or f"Pre CTI Data", filename="pre_cti_data"
                ),
            )

        cmap_original = self.mat_plot_2d.cmap

        if self.residuals_symmetric_cmap:
            symmetric_value = conf.instance["visualize"]["general"]["general"][
                "symmetric_cmap_value"
            ]

            self.mat_plot_2d.cmap = self.mat_plot_2d.cmap.symmetric_cmap_from(
                symmetric_value=symmetric_value
            )

        if pre_cti_data_residual_map:
            self.mat_plot_2d.plot_array(
                array=self.dataset.pre_cti_data_residual_map,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(
                    title=title_str or f"Pre CTI Data Residual Map",
                    filename="pre_cti_data_residual_map",
                ),
            )

        self.mat_plot_2d.cmap = cmap_original

        if cosmic_ray_map:
            self.mat_plot_2d.plot_array(
                array=self.dataset.cosmic_ray_map,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(
                    title=title_str or f"Cosmic Ray Map", filename="cosmic_ray_map"
                ),
            )

    def figures_1d(
        self,
        region: str,
        data: bool = False,
        data_logy: bool = False,
        noise_map: bool = False,
        pre_cti_data: bool = False,
        signal_to_noise_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `ImagingCI` object in 2D.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        if a `region` string is input, the 1D plots correspond to a region in 2D on the charge injection image, which
        is binned up over the parallel or serial direction to produce a 1D plot. For example, for the
        input `region=parallel_fpr`, this function extracts the FPR over each charge injection region and bins such
        that the 1D plot shows the FPR in the parallel direction.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        region
            The region on the charge injection image where data is extracted and binned over the parallel or serial
            direction {"parallel_fpr", "parallel_eper", "serial_fpr", "serial_eper"}
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars.
        data_logy
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale.
        noise_map
            Whether to make a 1D plot (via `plot`) of the noise-map extracted and binned over the region.
        pre_cti_data
            Whether to make a 1D plot (via `plot`) of the pre-cti data extracted and binned over the region.
        signal_to_noise_map
            Whether to make a 1D plot (via `plot`) of the signal-to-noise map data extracted and binned over
            the region.
        """

        if region == "fpr_non_uniformity":
            ls_errorbar = "-"
        else:
            ls_errorbar = ""

        should_plot_zero = self.should_plot_zero_from(region=region)

        title_str = self.title_str_from(region=region)

        if data:
            y = self.extract_region_from(array=self.dataset.data, region=region)
            y_errors = self.extract_region_noise_map_from(
                array=self.dataset.noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                ls_errorbar=ls_errorbar,
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Data {title_str}",
                    yunit="e-",
                    filename=f"data_{region}",
                ),
            )

        if data_logy:
            y = self.extract_region_from(array=self.dataset.data, region=region)
            y_errors = self.extract_region_noise_map_from(
                array=self.dataset.noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Data {title_str} [log10 y]",
                    yunit="e-",
                    filename=f"data_logy_{region}",
                ),
            )

        if noise_map:
            y = self.extract_region_from(array=self.dataset.noise_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Noise Map {title_str}",
                    ylabel="Noise (e-)",
                    filename=f"noise_map_{region}",
                ),
            )

        if pre_cti_data:
            y = self.extract_region_from(array=self.dataset.pre_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"CI Pre CTI {title_str}",
                    yunit="e-",
                    filename=f"pre_cti_data_{region}",
                ),
            )

        if signal_to_noise_map:
            y = self.extract_region_from(
                array=self.dataset.signal_to_noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Signal To Noise Map {title_str}",
                    ylabel="Signal To Noise (e-)",
                    filename=f"signal_to_noise_map_{region}",
                ),
            )

    def subplot(
        self,
        data: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        cosmic_ray_map: bool = False,
        auto_filename="subplot_dataset",
    ):
        """
        Plots the individual attributes of the plotter's `ImagingCI` object in 2D on a subplot.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is included on the subplot.

        Parameters
        ----------
        data
            Whether to include a 2D plot (via `imshow`) of the image data.
        noise_map
            Whether to include a 2D plot (via `imshow`) of the noise map.
        signal_to_noise_map
            Whether to include a 2D plot (via `imshow`) of the signal-to-noise map.
        pre_cti_data
            Whether to include a 2D plot (via `imshow`) of the pre-cti data.
        cosmic_ray_map
            Whether to include a 2D plot (via `imshow`) of the cosmic ray map.
        auto_filename
            The default filename of the output subplot if written to hard-disk.
        """
        self._subplot_custom_plot(
            data=data,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            pre_cti_data=pre_cti_data,
            cosmic_ray_map=cosmic_ray_map,
            auto_labels=AutoLabels(filename=auto_filename),
        )

    def subplot_dataset(self):
        """
        Standard subplot of the attributes of the plotter's `ImagingCI` object.
        """
        self.subplot(
            data=True,
            noise_map=True,
            signal_to_noise_map=True,
            pre_cti_data=True,
            cosmic_ray_map=True,
        )

    def subplot_1d(self, region: str):
        """
        Plots the individual attributes of the plotter's `ImagingCI` object in 1D on a subplot.

        These 1D plots correspond to a region in 2D on the charge injection image, which is binned up over the parallel
        or serial direction to produce a 1D plot. For example, for the input `region=parallel_fpr`, this
        function extracts the FPR over each charge injection region and bins such that the 1D plot shows the FPR
        in the parallel direction.

        The function plots the image, noise map, pre-cti data and signal to noise map in 1D on the subplot.

        Parameters
        ----------
        region
            The region on the charge injection image where data is extracted and binned over the parallel or serial
            direction {"parallel_fpr", "parallel_eper", "serial_fpr", "serial_eper"}
        """

        self.open_subplot_figure(number_subplots=4)

        self.figures_1d(data=True, region=region)
        self.figures_1d(noise_map=True, region=region)
        self.figures_1d(pre_cti_data=True, region=region)
        self.figures_1d(signal_to_noise_map=True, region=region)

        self.mat_plot_1d.output.subplot_to_figure(
            auto_filename=f"subplot_1d_ci_{region}"
        )
        self.close_subplot_figure()



import logging

from autoarray.layout.layout import Layout2D
from autoarray.layout.region import Region2D

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel("INFO")


class Layout2DACS(Layout2D):
    @classmethod
    def from_sizes(cls, roe_corner, serial_prescan_size=24, parallel_overscan_size=20):
        """
        Use an input array of the left quadrant in electrons and perform the rotations required to give correct
        arctic clocking.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.
        """

        parallel_overscan = Region2D(
            (2068 - parallel_overscan_size, 2068, serial_prescan_size, 2072)
        )

        serial_prescan = Region2D((0, 2068, 0, serial_prescan_size))

        return Layout2D.rotated_from_roe_corner(
            roe_corner=roe_corner,
            shape_native=(2068, 2072),
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
        )



from typing import Optional


class HyperCINoiseScalar(float):
    def __new__(cls, scale_factor=0.0):
        return super().__new__(cls, scale_factor)

    def __init__(self, scale_factor=0.0):
        """
        The hyper_ci-parameter factor by which the noises is scaled when included in the model-fitting process.
        """
        float.__init__(scale_factor)

    def scaled_noise_map_from(self, noise_scaling):
        """
        Returns the scaled noises map, by multiplying the noises-scaling image by the hyper_ci-parameter factor.
        """
        return self * noise_scaling


class HyperCINoiseCollection:
    def __init__(
        self,
        regions_ci: Optional[HyperCINoiseScalar] = None,
        parallel_eper: Optional[HyperCINoiseScalar] = None,
        serial_eper: Optional[HyperCINoiseScalar] = None,
        serial_overscan_no_eper: Optional[HyperCINoiseScalar] = None,
    ):
        self.regions_ci = regions_ci
        self.parallel_eper = parallel_eper
        self.serial_eper = serial_eper
        self.serial_overscan_no_eper = serial_overscan_no_eper

    @property
    def as_dict(self):
        hyper_dict = {
            "regions_ci": self.regions_ci,
            "parallel_eper": self.parallel_eper,
            "serial_eper": self.serial_eper,
            "serial_overscan_no_eper": self.serial_overscan_no_eper,
        }

        return {key: value for key, value in hyper_dict.items() if value is not None}



from functools import partial
from typing import List

import autofit as af
import autoarray as aa

from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D


def _dataset_1d_list_from(
    fit: af.Fit, use_dataset_full: bool = False
) -> List[Dataset1D]:
    """
    Returns a list of `Dataset1D` object from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The 1D dataset data as a .fits file (`dataset/data.fits`).
    - The noise-map as a .fits file (`dataset/noise_map.fits`).
    - The pre CTI data as a .fits file (`dataset/pre_cti_data.fits`).
    - The layout of the `Dataset1D` data structure used in the fit (`dataset/layout.json`).
    - The mask used to mask the `Dataset1D` data structure in the fit (`dataset/mask.fits`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a `Dataset1D` object, has the mask applied to the
    `Dataset1D` data structure and its settings updated to the values used by the model-fit.

    If multiple `Dataset1D` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
    is instead used to load lists of the data, noise-map, pre CTI data, layout and mask and combine them into a list of
    `Dataset1D` objects.

    If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
    to the database, the input `use_dataset_full` can be switched in to load instead the full `Dataset1D` objects.

    Parameters
    ----------
    fit
        A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
    use_dataset_full
        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `Dataset1D` objects.
    """

    if use_dataset_full:
        folder = "dataset_full"
    else:
        folder = "dataset"

    fit_list = [fit] if not fit.children else fit.children

    dataset_list = []

    for fit in fit_list:
        layout = fit.value(name=f"{folder}.layout")

        data = aa.Array1D.from_primary_hdu(primary_hdu=fit.value(name=f"{folder}.data"))
        noise_map = aa.Array1D.from_primary_hdu(
            primary_hdu=fit.value(name=f"{folder}.noise_map")
        )
        pre_cti_data = aa.Array1D.from_primary_hdu(
            primary_hdu=fit.value(name=f"{folder}.pre_cti_data")
        )

        dataset = Dataset1D(
            data=data,
            noise_map=noise_map,
            pre_cti_data=pre_cti_data,
            layout=layout,
        )

        mask = aa.Mask1D.from_primary_hdu(primary_hdu=fit.value(name=f"{folder}.mask"))

        dataset_list.append(dataset.apply_mask(mask=mask))

    return dataset_list


class Dataset1DAgg:
    def __init__(self, aggregator: af.Aggregator, use_dataset_full: bool = False):
        """
        Interfaces with an `PyAutoFit` aggregator object to create instances of `Dataset1D` objects from the results
        of a model-fit.

        The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

        - The 1D dataset data as a .fits file (`dataset/data.fits`).
        - The noise-map as a .fits file (`dataset/noise_map.fits`).
        - The pre CTI data as a .fits file (`dataset/pre_cti_data.fits`).
        - The layout of the `Dataset1D` data structure used in the fit (`dataset/layout.json`).
        - The mask used to mask the `Dataset1D` data structure in the fit (`dataset/mask.fits`).

        The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
        can load them all at once and create an `Dataset1D` object via the `_dataset_1d_from` method.

        This class's methods returns generators which create the instances of the `Dataset1D` objects. This ensures
        that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
        `Dataset1D` instances in the memory at once.

        For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
        creates instances of the corresponding 3 `Dataset1D` objects.

        If multiple `Dataset1D` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
        is instead used to load lists of the data, noise-map, pre CTI data, layout and mask and combine them into a list of
        `Dataset1D` objects.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `Dataset1D` objects.

        This can be done manually, but this object provides a more concise API.

        Parameters
        ----------
        aggregator
            A `PyAutoFit` aggregator object which can load the results of model-fits.
        use_dataset_full
            If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore
            accessible to the database, the input `use_dataset_full` can be switched in to load instead the
            full `Dataset1D` objects.
        """
        self.aggregator = aggregator
        self.use_dataset_full = use_dataset_full

    def dataset_list_gen_from(
        self,
    ):
        """
        Returns a generator of `Dataset1D` objects from an input aggregator.

        See `__init__` for a description of how the `Dataset1D` objects are created by this method.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `Dataset1D` objects.
        """

        func = partial(_dataset_1d_list_from, use_dataset_full=self.use_dataset_full)
        return self.aggregator.map(func=func)



from typing import Optional

from autoarray.plot.wrap.base.abstract import set_backend

set_backend()

import autoarray.plot as aplt

from autoarray.plot.abstract_plotters import AbstractPlotter

from autocti.plot.get_visuals.one_d import GetVisuals1D
from autocti.plot.get_visuals.two_d import GetVisuals2D


class Plotter(AbstractPlotter):
    def __init__(
        self,
        dataset,
        mat_plot_2d: aplt.MatPlot2D = None,
        visuals_2d: aplt.Visuals2D = None,
        include_2d: aplt.Include2D = None,
        mat_plot_1d: aplt.MatPlot1D = None,
        visuals_1d: aplt.Visuals1D = None,
        include_1d: aplt.Include1D = None,
    ):
        self.dataset = dataset

        super().__init__(
            mat_plot_2d=mat_plot_2d,
            visuals_2d=visuals_2d,
            include_2d=include_2d,
            mat_plot_1d=mat_plot_1d,
            visuals_1d=visuals_1d,
            include_1d=include_1d,
        )

    @property
    def get_1d(self):
        return GetVisuals1D(visuals=self.visuals_1d, include=self.include_1d)

    @property
    def get_2d(self):
        return GetVisuals2D(visuals=self.visuals_2d, include=self.include_2d)

    def title_str_from(self, region: Optional[str]) -> str:
        if self.dataset.settings_dict is not None:
            ccd_str = self.dataset.settings_dict.get("CCD")
        else:
            ccd_str = None

        if region is None:
            title_str = ""
        elif region == "fpr":
            title_str = "FPR"
        elif region == "eper":
            title_str = "EPER"
        elif region == "parallel_fpr":
            title_str = "Parallel FPR"
        elif region == "parallel_eper":
            title_str = "Parallel EPER"
        elif region == "serial_fpr":
            title_str = "Serial FPR"
        elif region == "serial_eper":
            title_str = "Serial EPER"

        if ccd_str is None:
            return title_str
        return f"{ccd_str} {title_str}"

    def title_str_2d_from(self) -> Optional[str]:
        if self.dataset.settings_dict is not None:
            ccd_str = self.dataset.settings_dict.get("CCD")
            ig1_str = self.dataset.settings_dict.get("CI_IG1")
            ig2_str = self.dataset.settings_dict.get("CI_IG2")
            id_delay_str = self.dataset.settings_dict.get("CI_IDDLY")

            return f"{ccd_str} IG1={ig1_str} IG2={ig2_str} IDD={id_delay_str}"

    def text_manual_dict_from(self, region: str):
        try:
            fpr_value = self.dataset.fpr_value
        except AttributeError:
            fpr_value = None

        text_manual_dict = {}

        if region is not None:
            if fpr_value is not None and "eper" in region:
                fpr_dict = {"FPR (e-)": self.dataset.fpr_value}
                text_manual_dict = {**text_manual_dict, **fpr_dict}

        if self.dataset.settings_dict is not None:
            text_manual_dict = {**text_manual_dict, **self.dataset.settings_dict}

        return text_manual_dict

    def text_manual_dict_y_from(self, region: str):
        if region is None or "eper" in region:
            return 0.94
        return 0.34

    def should_plot_zero_from(self, region: Optional[str]):
        if region is None:
            return False

        if "eper" in region:
            return True

        return False



from typing import Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.serial.abstract import Extract2DSerial
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DSerialFPR(Extract2DSerial):
    def region_list_from(self, settings: SettingsExtract):
        """
        Returns a list of the 2D serial FPR regions from the `region_list` containing signal  (e.g. the charge
        injection regions of charge injection data), between two input `pixels` indexes.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the serial FPRs (e.g.
        the serial prescan) are extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sss]
           [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        |  [...][1c1c1cc1c1cc1ccc1cc1c][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the first serial FPR all charge injection regions:

        region_list[0] = [0, 2, 3, 21] (serial prescan is 3 pixels)
        region_list[1] = [4, 6, 3, 21] (serial prescan is 3 pixels)

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        serial FPR of each charge injection region:

        array_2d_list[0] =[c0c0]
        array_2d_list[1] =[1c1c]

        Parameters
        ----------
        settings
            The settings used to extract the serial region FPRs from, which for example include the `pixels`
            tuple specifying the range of pixel columns they are extracted between.
        """
        return [
            region.serial_front_region_from(
                pixels=settings.pixels, pixels_from_end=settings.pixels_from_end
            )
            for region in self.region_list
        ]

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        """
        The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
        CTI modeling in 1D.

        This is performed by binning up the data via the `binned_array_1d_from` function.

        In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
        charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

        The function returns the this region if the 1D dataset is extracted from the serial FPRs. This is
        the full range of the `pixels` tuple, unless negative entries are included, meaning that pixels
        before the FPRs are extracted.

        Parameters
        ----------
        settings
            The settings used to extract the serial region FPRs from, which for example include the `pixels`
            tuple specifying the range of pixel columns they are extracted between.
        """
        return extract_2d_util.binned_region_1d_fpr_from(pixels=settings.pixels)



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from autocti.clocker.abstract import AbstractClocker
    from autocti.model.model_util import CTI1D
    from autocti.model.model_util import CTI2D
    from autocti.dataset_1d.fit import FitDataset1D

import autofit as af

from autocti.aggregator.abstract import AbstractAgg
from autocti.aggregator.dataset_1d import _dataset_1d_list_from


def _fit_dataset_1d_list_from(
    fit: af.Fit,
    cti: Union[CTI1D, CTI2D],
    use_dataset_full: bool = False,
    clocker_list: Optional[AbstractClocker] = None,
) -> List[FitDataset1D]:
    """
    Returns a list of `FitDataset1D` object from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The masked dataset (e.g. data / noise map / pre cti data) as .fits files (contained in `dataset` folder).
    - The clocker used to add  CTI in the fit (`dataset/clocker.json`).
    - The settings used for clocking CIT (contained in `dataset/settings_cti.json`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a  list of `FitDataset1D` objects, by loading the masked
    dataset adding CTI to its pre-cti data via the cti model and clocking and fitting the model image to the dataset.

    If multiple `Dataset1D` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
    is instead used to load lists of the datasets, perform the fit and return a list of `FitDataset1D` objects.

    If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
    to the database, the input `use_dataset_full` can be switched in to fit the full dataset instead.

    Parameters
    ----------
    fit
        A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
    use_dataset_full
        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `Dataset1D` objects.
    clocker_list
        If input, overwrites the clocker used in the fit with a new clocker which is used to perform the fit.
    """

    from autocti.dataset_1d.fit import FitDataset1D

    dataset_list = _dataset_1d_list_from(fit=fit, use_dataset_full=use_dataset_full)

    if clocker_list is None:
        if not fit.children:
            clocker_list = [fit.value(name="clocker")]
        else:
            clocker_list = fit.child_values(name="clocker")

    post_cti_data_list = [
        clocker.add_cti(data=dataset.pre_cti_data, cti=cti)
        for dataset, clocker in zip(dataset_list, clocker_list)
    ]

    return [
        FitDataset1D(
            dataset=dataset,
            post_cti_data=post_cti_data,
        )
        for dataset, post_cti_data in zip(dataset_list, post_cti_data_list)
    ]


class FitDataset1DAgg(AbstractAgg):
    def __init__(
        self,
        aggregator: af.Aggregator,
        use_dataset_full: bool = False,
        clocker_list: Optional[List[AbstractClocker]] = None,
    ):
        """
        Interfaces with an `PyAutoFit` aggregator object to create instances of `Dataset1D` objects from the results
        of a model-fit.

        The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

        - The masked dataset (e.g. data / noise map / pre cti data) as .fits files (contained in `dataset` folder).
        - The clocker used to add  CTI in the fit (`dataset/clocker.json`).
        - The settings used for clocking CIT (contained in `dataset/settings_cti.json`).

        The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
        can load them all at once and create a `FitDataset1D` object via the `_fit_dataset_1d_from` method.

        This class's methods returns generators which create the instances of the `FitDataset1D` objects. This ensures
        that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
        `Dataset1D` instances in the memory at once.

        For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
        creates instances of the corresponding 3 `Dataset1D` objects.

        If multiple `Dataset1D` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
        is instead used to load lists of the datasets, perform the fit and return a list of `FitDataset1D` objects.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to fit the full dataset instead.

        This can be done manually, but this object provides a more concise API.

        Parameters
        ----------
        aggregator
            A `PyAutoFit` aggregator object which can load the results of model-fits.
        use_dataset_full
            If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore
            accessible to the database, the input `use_dataset_full` can be switched in to load instead the
            full `Dataset1D` objects.
        clocker_list
            If input, overwrites the clocker used in the fit with a new clocker which is used to perform the fit.
        """
        super().__init__(
            aggregator=aggregator,
            use_dataset_full=use_dataset_full,
            clocker_list=clocker_list,
        )

    def object_via_gen_from(self, fit, cti: Union[CTI1D, CTI2D]) -> List[FitDataset1D]:
        """
        Returns a generator of `FitDataset1D` objects from an input aggregator.

        See `__init__` for a description of how the `FitDataset1D` objects are created by this method.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to fit the full dataset instead.

        Parameters
        ----------
        fit
            A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
        cti
            The CTI model used to add CTI to the dataset to perform the fit.
        """
        return _fit_dataset_1d_list_from(
            fit=fit,
            cti=cti,
            use_dataset_full=self.use_dataset_full,
            clocker_list=self.clocker_list,
        )



from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.one_d.abstract import Extract1D
from autocti.extract.settings import SettingsExtract


class Extract1DFPR(Extract1D):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region1D]:
        """
        Returns a list of the (x0, x1) regions containing the FPRs of a 1D CTI dataset.

        These are used for extracting the FPR regions of 1D data.

        Negative pixel values can be input into the `pixels` tuple, whereby pixels in front of the FPRs  are also
        extracted.

        Parameters
        ----------
        pixels
            The row indexes to extract the front edge between (e.g. pixels(0, 3) extracts the 1st, 2nd and 3rd pixels)
        """
        return list(
            map(
                lambda region: region.front_region_from(
                    pixels=settings.pixels, pixels_from_end=settings.pixels_from_end
                ),
                self.region_list,
            )
        )



import numpy as np
from typing import Optional, Tuple

import autoarray as aa

from autocti.layout.one_d import Layout1D
from autocti.extract.settings import SettingsExtract


class SettingsMask1D:
    def __init__(
        self,
        fpr_pixels: Optional[Tuple[int, int]] = None,
        eper_pixels: Optional[Tuple[int, int]] = None,
    ):
        self.fpr_pixels = fpr_pixels
        self.eper_pixels = eper_pixels


class Mask1D(aa.Mask1D):
    @classmethod
    def masked_fpr_and_eper_from(
        cls,
        mask: "Mask1D",
        layout: Layout1D,
        settings: "SettingsMask1D",
        pixel_scales: aa.type.PixelScales,
    ) -> "Mask1D":
        if settings.fpr_pixels is not None:
            fpr_mask = cls.masked_fpr_from_layout(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + fpr_mask

        if settings.eper_pixels is not None:
            eper_mask = cls.masked_eper_from_layout(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + eper_mask

        return mask

    @classmethod
    def masked_fpr_from_layout(
        cls,
        layout: Layout1D,
        settings: "SettingsMask1D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask1D":
        fpr_regions = layout.extract.fpr.region_list_from(
            settings=SettingsExtract(pixels=settings.fpr_pixels)
        )

        mask = np.full(layout.shape_1d, False)

        for region in fpr_regions:
            mask[region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask1D(mask=mask.astype("bool"), pixel_scales=pixel_scales)

    @classmethod
    def masked_eper_from_layout(
        cls,
        layout: Layout1D,
        settings: "SettingsMask1D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask1D":
        eper_regions = layout.extract.eper.region_list_from(
            settings=SettingsExtract(pixels=settings.eper_pixels)
        )

        mask = np.full(layout.shape_1d, False)

        for region in eper_regions:
            mask[region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask1D(mask=mask.astype("bool"), pixel_scales=pixel_scales)



from typing import List

import autoarray as aa

from autocti.extract.two_d.parallel.abstract import Extract2DParallel
from autocti.extract.settings import SettingsExtract


class Extract2DParallelPedestal(Extract2DParallel):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D pedestral region, which is in the corner of the CCD and extracted using
        the coordinates of the parallel and serial overscan regions.

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        The parallel pedestal spans all rows of the parallel overscan and columns of the serial overscan. Unlike
        the parallel and serial overscans, the pedestal is not directional. Separatel parallel and serial pedestal
        extractor objects are used to extract the pedestal in theses directions.

        Negative pixel values can be input into the `pixels` tuple, whereby rows in front of the pedestal are
        extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [pepepepepe] = pedestal
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp][pepe]
               [ppppppppppppppppppppp][pepe]
           [...][ttttttttttttttttttttt][sss]
           [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        |  [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        \/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/

         []     [=====================]
                <---------Ser--------

         The extracted regions correspond to the parallel overscan [pepe] regions.

         For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
         pedestal pixels across the rows of the overscan:

         array_2d_list[0] = [pepe]

         Parameters
         ----------
        settings
           The settings used to extract the parallel pedestal, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """
        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.pedestal.total_rows - settings.pixels_from_end,
                self.pedestal.total_rows,
            )

        pedestal_extract = aa.Region2D(
            (
                self.pedestal.y0 + pixels[0],
                self.pedestal.y0 + pixels[1],
                self.pedestal.x0,
                self.pedestal.x1,
            )
        )

        return [pedestal_extract]

    def array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
        Extract from an `Array2D` the pedestal and return it as a new `Array2D` where everything else (e.g.
        the charge injection regions, serial EPERs) are zeros.

        The dimensions of the input array therefore do not change (unlike other ``Extract`` methods).

        Negative pixel values can be input into the `pixels` tuple, whereby rows in front of the parallel pedestal (e.g.
        the serial EPERs) are extracted.

        The diagram below illustrates the extraction:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [pepepepepe] = pedestal
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

             [tptpptptptpptpptpptpt][pepe]
             [tptptptpptpttptptptpt][pepe]
        [...][ttttttttttttttttttttt][sss]
        [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        Par[...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
        |/ [...][ccccccccccccccccccccc][sss]   \/

        []     [=====================]
             <--------Ser---------

        The extracted array keeps only the pedestal, everything else become 0s:

              [0000000000000000000000][pepe]
              [0000000000000000000000][pepe]
          [000][000000000000000000000][000]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][000000000000000000000][000]    | Direction
        Par[000][00000000000000000000][000]    | of
        | [000][000000000000000000000][000]    | clocking
        |/ [000][000000000000000000000][000]   \/

        []     [=====================]
             <--------Ser---------
        """

        array_2d = array.native.copy() * 0.0

        return self.add_to_array(
            new_array=array_2d,
            array=array,
            settings=SettingsExtract(pixels=(0, self.pedestal.total_rows)),
        )



import numpy as np
from typing import Optional

from autoarray.structures.arrays.uniform_2d import Array2D

from autocti.instruments.euclid.header import HeaderEuclid

from autoarray.layout import layout_util


class Array2DEuclid(Array2D):
    """
    In the Euclid FPA, the quadrant id ('E', 'F', 'G', 'H') depends on whether the CCD is located
    on the left side (rows 1-3) or right side (rows 4-6) of the FPA:

    LEFT SIDE ROWS 1-2-3
    --------------------

     <--------S-----------   ---------S----------->
    [] [========= 2 =========] [========= 3 =========] []          |
    /    [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  /          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | Direction arctic
    P   [xxxxxxxxx H xxxxxxxxx] [xxxxxxxxx G xxxxxxxxx]  P         | clocks an image
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | without any rotation
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | (e.g. towards row 0
                                                                   | of the ndarrays)
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
    P   [xxxxxxxxx E xxxxxxxxx] [xxxxxxxxx F xxxxxxxxx] P          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]            |

    [] [========= 0 =========] [========= 1 =========] []
        <---------S----------   ----------S----------->


    RIGHT SIDE ROWS 4-5-6
    ---------------------

     <--------S-----------   ---------S----------->
    [] [========= 2 =========] [========= 3 =========] []          |
    /    [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  /          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | Direction arctic
    P   [xxxxxxxxx F xxxxxxxxx] [xxxxxxxxx E xxxxxxxxx]  P         | clocks an image
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | without any rotation
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | (e.g. towards row 0
                                                                   | of the ndarrays)
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
    P   [xxxxxxxxx G xxxxxxxxx] [xxxxxxxxx H xxxxxxxxx] P          |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]            |

    [] [========= 0 =========] [========= 1 =========] []
        <---------S----------   ----------S----------->

    Therefore, to setup a quadrant image with the correct layout using its CCD id (from which
    we can extract its row number) and quadrant id, we need to first determine if the CCD is on the left / right
    side and then use its quadrant id ('E', 'F', 'G' or 'H') to pick the correct quadrant.
    """

    @classmethod
    def from_fits_header(cls, array, ext_header):
        """
        Use an input array of a Euclid quadrant and its corresponding .fits file header to rotate the quadrant to
        the correct orientation for arCTIc clocking.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.
        """

        ccd_id = ext_header["CCDID"]
        quadrant_id = ext_header["QUADID"]

        return cls.from_ccd_and_quadrant_id(
            array=array, ccd_id=ccd_id, quadrant_id=quadrant_id
        )

    @classmethod
    def from_ccd_and_quadrant_id(cls, array, ccd_id, quadrant_id):
        """
        Use an input array of a Euclid quadrant, its ccd_id and quadrant_id  to rotate the quadrant to
        the correct orientation for arCTIc clocking.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.
        """

        row_index = ccd_id[-1]

        if (row_index in "123") and (quadrant_id == "E"):
            return Array2DEuclid.bottom_left(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "123") and (quadrant_id == "F"):
            return Array2DEuclid.bottom_right(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "123") and (quadrant_id == "G"):
            return Array2DEuclid.top_right(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "123") and (quadrant_id == "H"):
            return Array2DEuclid.top_left(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "456") and (quadrant_id == "E"):
            return Array2DEuclid.top_right(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "456") and (quadrant_id == "F"):
            return Array2DEuclid.top_left(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "456") and (quadrant_id == "G"):
            return Array2DEuclid.bottom_left(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )
        elif (row_index in "456") and (quadrant_id == "H"):
            return Array2DEuclid.bottom_right(
                array_electrons=array, ccd_id=ccd_id, quadrant_id=quadrant_id
            )

    @classmethod
    def top_left(
        cls,
        array_electrons: np.ndarray,
        ccd_id: Optional[str] = None,
        quadrant_id: Optional[str] = None,
    ) -> "Array2DEuclid":
        """
        Load a Euclid quadrant where the read out electronics (ROE) are at the top left of the CCD.

        The array is loaded and then rotated to ensure that parallel and serial CTI clocking go in the correct
        direction for arctic clocking.

        ROE's in the `top_left` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id H.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id F.

        The data is rotated by a flip-up-down, such that the ROE electrons are in the bottom left corner after rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        array_electrons
            The array of data (in units of electrons) which has the ROE in the top left and is rotated for arctic
            clocking.
        ccd_id
            The ID of the euclid CCD (1, 2, 3, 4, 5, or 6) indicating which half of the FP the CCD is in.
        quadrant_id
            The ID of the quadrant (E, G, F or H).
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(0, 0)
        )

        header = HeaderEuclid(
            original_roe_corner=(0, 0), ccd_id=ccd_id, quadrant_id=quadrant_id
        )

        return cls.no_mask(values=array_electrons, pixel_scales=0.1, header=header)

    @classmethod
    def top_right(
        cls,
        array_electrons: np.ndarray,
        ccd_id: Optional[str] = None,
        quadrant_id: Optional[str] = None,
    ) -> "Array2DEuclid":
        """
        Load a Euclid quadrant where the read out electronics (ROE) are at the top right of the CCD.

        The array is loaded and then rotated to ensure that parallel and serial CTI clocking go in the correct
        direction for arctic clocking.

        ROE's in the `top_right` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id G.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id E.

        The data is rotated by a flip-up-down and flip left-right, such that the ROE electrons are in the bottom left
        corner after rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        array_electrons
            The array of data (in units of electrons) which has the ROE in the top right and is rotated for arctic
            clocking.
        ccd_id
            The ID of the euclid CCD (1, 2, 3, 4, 5, or 6) indicating which half of the FP the CCD is in.
        quadrant_id
            The ID of the quadrant (E, G, F or H).
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(0, 1)
        )

        header = HeaderEuclid(
            original_roe_corner=(0, 1), ccd_id=ccd_id, quadrant_id=quadrant_id
        )

        return cls.no_mask(values=array_electrons, pixel_scales=0.1, header=header)

    @classmethod
    def bottom_left(
        cls,
        array_electrons: np.ndarray,
        ccd_id: Optional[str] = None,
        quadrant_id: Optional[str] = None,
    ) -> "Array2DEuclid":
        """
        Load a Euclid quadrant where the read out electronics (ROE) are at the bottom left of the CCD.

        The array is loaded and then rotated to ensure that parallel and serial CTI clocking go in the correct
        direction for arctic clocking.

        ROE's in the `bottom left` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id E.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id G.

        No rotation is performed for arCTIc clocking because the code assume the ROE are at the bottom left.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        array_electrons
            The array of data (in units of electrons) which has the ROE in the bottom left and is rotated for arctic
            clocking.
        ccd_id
            The ID of the euclid CCD (1, 2, 3, 4, 5, or 6) indicating which half of the FP the CCD is in.
        quadrant_id
            The ID of the quadrant (E, G, F or H).
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(1, 0)
        )

        header = HeaderEuclid(
            original_roe_corner=(1, 0), ccd_id=ccd_id, quadrant_id=quadrant_id
        )

        return cls.no_mask(values=array_electrons, pixel_scales=0.1, header=header)

    @classmethod
    def bottom_right(
        cls,
        array_electrons: np.ndarray,
        ccd_id: Optional[str] = None,
        quadrant_id: Optional[str] = None,
    ) -> "Array2DEuclid":
        """
        Load a Euclid quadrant where the read out electronics (ROE) are at the bottom right of the CCD.

        The array is loaded and then rotated to ensure that parallel and serial CTI clocking go in the correct
        direction for arctic clocking.

        ROE's in the `bottom right` correspond to 1/4 of the 144 quadrants on a Euclid FPA, specifically:

        - Those in the top half of the FPA (CCDID = 1, 2 or 3) and quadrant id F.
        - Those in the bottom half of the FPA (CCDID = 4, 5 or 6) and quadrant id H.

        The data is rotated by a flip left-right, such that the ROE electrons are in the bottom left corner after
        rotation.

        See the docstring of the `Array2DEuclid` class for a complete description of the Euclid FPA, quadrants and
        rotations.

        Parameters
        ----------
        array_electrons
            The array of data (in units of electrons) which has the ROE in the bottom right and is rotated for arctic
            clocking.
        ccd_id
            The ID of the euclid CCD (1, 2, 3, 4, 5, or 6) indicating which half of the FP the CCD is in.
        quadrant_id
            The ID of the quadrant (E, G, F or H).
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(1, 1)
        )

        header = HeaderEuclid(
            original_roe_corner=(1, 1), ccd_id=ccd_id, quadrant_id=quadrant_id
        )

        return cls.no_mask(values=array_electrons, pixel_scales=0.1, header=header)



from autoarray.geometry import geometry_util as geometry
from autoarray.mask import mask_1d_util as mask_1d
from autoarray.mask import mask_2d_util as mask_2d
from autoarray.structures.arrays import array_1d_util as array_1d
from autoarray.structures.arrays import array_2d_util as array_2d
from autoarray.structures.grids import grid_1d_util as grid_1d
from autoarray.structures.grids import grid_2d_util as grid_2d
from autoarray.structures.grids import sparse_2d_util as sparse
from autoarray.layout import layout_util as layout
from autoarray.fit import fit_util as fit
from autocti.model import model_util as model
from autocti.extract.two_d import extract_2d_util as extract_2d
from autocti.charge_injection import ci_util as ci

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)



import autoarray.plot as aplt

from autocti.charge_injection.fit import FitImagingCI


class GetVisuals2D(aplt.GetVisuals2D):
    def __init__(self, include: aplt.Include2D, visuals: aplt.Visuals2D):
        """
        Class which gets 2D attributes and adds them to a `Visuals2D` objects, such that they are plotted on 2D figures.

        For a visual to be extracted and added for plotting, it must have a `True` value in its corresponding entry in
        the `Include2D` object. If this entry is `False`, the `GetVisuals2D.get` method returns a None and the
        attribute is omitted from the plot.

        The `GetVisuals2D` class adds new visuals to a pre-existing `Visuals2D` object that is passed to
        its `__init__` method. This only adds a new entry if the visual are not already in this object.

        Parameters
        ----------
        include
            Sets which 2D visuals are included on the figure that is to be plotted (only entries which are `True`
            are extracted via the `GetVisuals2D` object).
        visuals
            The pre-existing visuals of the plotter which new visuals are added too via the `GetVisuals2D` class.
        """
        super().__init__(include=include, visuals=visuals)

    def via_fit_imaging_ci_from(self, fit: FitImagingCI) -> aplt.Visuals2D:
        """
        From a `FitImagingCI` get its attributes that can be plotted and return them in a `Visuals2D` object.

        Only attributes not already in `self.visuals` and with `True` entries in the `Include2D` object are extracted
        for plotting.

        From a `FitImagingCI` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the 2D coordinate system.
        - mask: the 2D mask.
        - border: the border of the 2D mask, which are all of the mask's exterior edge pixels.
        - parallel overscan: the 2D region defining the parallel overscan on the imaging data.
        - serial prescan: the 2D region defining the serial prerscan on the imaging data.
        - serial overscan: the 2D region defining the serial overscan on the imaging data.

        Parameters
        ----------
        fit
            The fit imaging object whose attributes are extracted for plotting.

        Returns
        -------
        Visuals2D
            The collection of attributes that are plotted by a `Plotter` object.
        """
        visuals_2d_via_fit = super().via_fit_imaging_from(fit=fit)

        parallel_overscan = (
            self.get("parallel_overscan", fit.dataset.layout.parallel_overscan),
        )
        serial_prescan = (
            self.get("serial_prescan", fit.dataset.layout.serial_prescan),
        )
        serial_overscan = (
            self.get("serial_overscan", fit.dataset.layout.serial_overscan),
        )

        return visuals_2d_via_fit + self.visuals.__class__(
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )



from astropy.io import fits
import copy
import logging
import numpy as np
from pathlib import Path
import os
from typing import Union

from autoarray import exc
from autoarray.layout import layout_util

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel("INFO")


def fits_hdu_via_quadrant_letter_from(quadrant_letter):
    if quadrant_letter == "D" or quadrant_letter == "C":
        return 1
    elif quadrant_letter == "B" or quadrant_letter == "A":
        return 4
    else:
        raise exc.ArrayException("Quadrant letter for FrameACS must be A, B, C or D.")


def array_eps_to_counts(array_eps, bscale, bzero):
    if bscale is None:
        raise exc.ArrayException(
            "Cannot convert a Frame2D to units COUNTS without a bscale attribute (bscale = None)."
        )

    return (array_eps - bzero) / bscale


def prescan_fitted_bias_column(prescan, n_rows=2048, n_rows_ov=20):
    """
    Generate a bias column to be subtracted from the main image by doing a
    least squares fit to the serial prescan region.

    e.g. image -= prescan_fitted_bias_column(image[18:24])

    See Anton & Rorres (2013), S9.3, p460.

    Parameters
    ----------
    prescan : [[float]]
        The serial prescan part of the image. Should usually cover the full
        number of rows but may skip the first few columns of the prescan to
        avoid trails.

    n_rows
        The number of rows in the image, exculding overscan.

    n_rows_ov, int
        The number of overscan rows in the image.

    Returns
    -------
    bias_column : [float]
        The fitted bias to be subtracted from all image columns.
    """
    n_columns_fit = prescan.shape[1]

    # Flatten the multiple fitting columns to a long 1D array
    # y = [y_1_1, y_2_1, ..., y_nrow_1, y_1_2, y_2_2, ..., y_nrow_ncolfit]
    y = prescan[:-n_rows_ov].T.flatten()
    # x = [1, 2, ..., nrow, 1, ..., nrow, 1, ..., nrow, ...]
    x = np.tile(np.arange(n_rows), n_columns_fit)

    # M = [[1, 1, ..., 1], [x_1, x_2, ..., x_n]].T
    M = np.array([np.ones(n_rows * n_columns_fit), x]).T

    # Best-fit values for y = M v
    v = np.dot(np.linalg.inv(np.dot(M.T, M)), np.dot(M.T, y))

    # Map to full image size for easy subtraction
    bias_column = v[0] + v[1] * np.arange(n_rows + n_rows_ov)

    return np.transpose([bias_column])


def output_quadrants_to_fits(
    file_path: Union[Path, str],
    quadrant_a,
    quadrant_b,
    quadrant_c,
    quadrant_d,
    header_a=None,
    header_b=None,
    header_c=None,
    header_d=None,
    overwrite: bool = False,
):
    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    array_hdu_1 = np.zeros((2068, 4144))
    array_hdu_4 = np.zeros((2068, 4144))

    def get_header(quadrant):
        try:
            return quadrant.header
        except AttributeError:
            raise (
                "You must pass in the header of the quadrants to output them to an ACS fits file."
            )

    header_a = get_header(quadrant_a) if header_a is None else header_a
    try:
        quadrant_a = copy.copy(np.asarray(quadrant_a.native))
    except AttributeError:
        quadrant_a = copy.copy(np.asarray(quadrant_a))

    quadrant_a = quadrant_convert_to_original(
        quadrant=quadrant_a, roe_corner=(1, 0), header=header_a, use_flipud=True
    )
    array_hdu_4[0:2068, 0:2072] = quadrant_a

    header_b = get_header(quadrant_b) if header_b is None else header_b

    try:
        quadrant_b = copy.copy(np.asarray(quadrant_b.native))
    except AttributeError:
        quadrant_b = copy.copy(np.asarray(quadrant_b))
    quadrant_b = quadrant_convert_to_original(
        quadrant=quadrant_b, roe_corner=(1, 1), header=header_b, use_flipud=True
    )
    array_hdu_4[0:2068, 2072:4144] = quadrant_b

    header_c = get_header(quadrant_c) if header_c is None else header_c
    try:
        quadrant_c = copy.copy(np.asarray(quadrant_c.native))
    except AttributeError:
        quadrant_c = copy.copy(np.asarray(quadrant_c))
    quadrant_c = quadrant_convert_to_original(
        quadrant=quadrant_c, roe_corner=(1, 0), header=header_c, use_flipud=False
    )
    array_hdu_1[0:2068, 0:2072] = quadrant_c

    header_d = get_header(quadrant_d) if header_d is None else header_d
    try:
        quadrant_d = copy.copy(np.asarray(quadrant_d.native))
    except AttributeError:
        quadrant_d = copy.copy(np.asarray(quadrant_d))
    quadrant_d = quadrant_convert_to_original(
        quadrant=quadrant_d, roe_corner=(1, 1), header=header_d, use_flipud=False
    )
    array_hdu_1[0:2068, 2072:4144] = quadrant_d

    hdu_list = fits.HDUList()

    hdu_list.append(fits.ImageHDU())
    hdu_list.append(fits.ImageHDU(array_hdu_1))
    hdu_list.append(fits.ImageHDU())
    hdu_list.append(fits.ImageHDU())
    hdu_list.append(fits.ImageHDU(array_hdu_4))
    hdu_list.append(fits.ImageHDU())

    def set_header(header):
        header.set("cticor", "ARCTIC", "CTI CORRECTION PERFORMED USING ARCTIC")
        return header

    hdu_list[0].header = set_header(header_a.header_sci_obj)
    hdu_list[1].header = set_header(header_c.header_hdu_obj)
    hdu_list[4].header = set_header(header_a.header_hdu_obj)
    hdu_list.writeto(file_path)


def quadrant_convert_to_original(
    quadrant, roe_corner, header, use_flipud=False, use_calibrated_gain=True
):
    if header.bias is not None:
        quadrant += header.bias.native

    if header.bias_serial_prescan_column is not None:
        quadrant += header.bias_serial_prescan_column

    quadrant = header.array_electrons_to_original(
        array=quadrant, use_calibrated_gain=use_calibrated_gain
    )

    if use_flipud:
        quadrant = np.flipud(quadrant)

    return layout_util.rotate_array_via_roe_corner_from(
        array=quadrant, roe_corner=roe_corner
    )



from typing import Tuple

import autoarray as aa


def binned_region_1d_fpr_from(pixels: Tuple[int, int]) -> aa.Region1D:
    """
    The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
    CTI modeling in 1D.

    This is performed by binning up the data via the `binned_array_1d_from` function.

    In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
    charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

    The function returns the this region if the 1D dataset is extracted from the FPRs. This is the full range of
    the `pixels` tuple, unless negative entries are included, meaning that pixels before the FPRs are extracted.

    Parameters
    ----------
    pixels
        The row / column pixel index to extract the FPRs between (e.g. `pixels=(0, 3)` extracts the 1st, 2nd and 3rd
        FPR rows / columns)
    """
    if pixels[0] <= 0 and pixels[1] <= 0:
        return None
    elif pixels[0] >= 0:
        return aa.Region1D(region=(0, pixels[1]))
    return aa.Region1D(region=(abs(pixels[0]), pixels[1] + abs(pixels[0])))


def binned_region_1d_eper_from(pixels: Tuple[int, int]) -> aa.Region1D:
    """
    `Extract` objects extract arrays and other values from a 2D CTI dataset, which is used to perform
    CTI modeling in 1D.

    This is performed by binning up the data via the `binned_array_1d_from` function.

    In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
    charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

    The function returns this region if the 1D dataset is extracted from the EPERs. The charge region is only included
    if there are negative entries in the `pixels` tuple, meaning that pixels before the EPERs (e.g. the FPR) are
    extracted.

    Parameters
    ----------
    pixels
        The row / column pixel index to extract the FPRs between (e.g. `pixels=(0, 3)` extracts the 1st, 2nd and 3rd
        FPR rows / columns)
    """
    if pixels[0] >= 0:
        return None
    elif pixels[1] >= 0:
        return aa.Region1D(region=(0, -pixels[0]))
    return aa.Region1D(region=(0, pixels[1] - pixels[0]))



from typing import Callable, Optional

import autoarray.plot as aplt

from autoarray.plot.auto_labels import AutoLabels

from autocti.plot.abstract_plotters import Plotter
from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D


class Dataset1DPlotter(Plotter):
    def __init__(
        self,
        dataset: Dataset1D,
        mat_plot_1d: aplt.MatPlot1D = aplt.MatPlot1D(),
        visuals_1d: aplt.Visuals1D = aplt.Visuals1D(),
        include_1d: aplt.Include1D = aplt.Include1D(),
    ):
        """
        Plots the attributes of `Dataset1D` objects using the matplotlib method `line()` and many other matplotlib
        functions which customize the plot's appearance.

        The `mat_plot_1d` attribute wraps matplotlib function calls to make the figure. By default, the settings
        passed to every matplotlib function called are those specified in the `config/visualize/mat_wrap/*.ini` files,
        but a user can manually input values into `MatPlot1d` to customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals1D` object. Attributes may be extracted from
        the `Imaging` and plotted via the visuals object, if the corresponding entry is `True` in the `Include1D`
        object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        dataset
            The dataset 1d the plotter plots.
        mat_plot_1d
            Contains objects which wrap the matplotlib function calls that make 1D plots.
        visuals_1d
            Contains 1D visuals that can be overlaid on 1D plots.
        include_1d
            Specifies which attributes of the `ImagingCI` are extracted and plotted as visuals for 1D plots.
        """
        super().__init__(
            dataset=dataset,
            mat_plot_1d=mat_plot_1d,
            include_1d=include_1d,
            visuals_1d=visuals_1d,
        )

    def get_visuals_1d(self) -> aplt.Visuals1D:
        return self.visuals_1d

    @property
    def extract_region_from(self) -> Callable:
        return self.dataset.layout.extract_region_from

    def figures_1d(
        self,
        region: Optional[str] = None,
        data: bool = False,
        data_logy: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `Dataset1D` object in 1D.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        If a `region` string is input, the 1D plots correspond to regions in 1D on the 1D dataset, which are binned up
        to produce a1D plot.

        For example, for the input `region=fpr`, this function extracts the FPR over each charge region and bins them
        such that the 1D plot shows the average FPR.

        The API is such that every plottable attribute of the `Dataset1D` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        region
            The region on the 1D dataset where data is extracted and binned {fpr", "eper"}
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars.
        data_logy
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale.
        noise_map
            Whether to make a 1D plot (via `plot`) of the noise-map extracted and binned over the region.
        pre_cti_data
            Whether to make a 1D plot (via `plot`) of the pre-cti data extracted and binned over the region.
        signal_to_noise_map
            Whether to make a 1D plot (via `plot`) of the signal-to-noise map data extracted and binned over
            the region.
        """

        suffix = f"_{region}" if region is not None else ""
        title_str = self.title_str_from(region=region)

        should_plot_zero = self.should_plot_zero_from(region=region)

        if data:
            y = self.extract_region_from(array=self.dataset.data, region=region)
            y_errors = self.extract_region_from(
                array=self.dataset.noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                visuals_1d=self.get_visuals_1d(),
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                auto_labels=AutoLabels(
                    title=f"Data 1D {title_str}",
                    yunit="e-",
                    filename=f"data{suffix}",
                ),
            )

        if data_logy:
            y = self.extract_region_from(array=self.dataset.data, region=region)
            y_errors = self.extract_region_from(
                array=self.dataset.noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                visuals_1d=self.get_visuals_1d(),
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                auto_labels=AutoLabels(
                    title=f"Data 1D {title_str} [log10]",
                    yunit="e-",
                    filename=f"data_logy{suffix}",
                ),
            )

        if noise_map:
            y = self.extract_region_from(array=self.dataset.noise_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Noise Map {title_str}",
                    yunit="e-",
                    filename=f"noise_map{suffix}",
                ),
            )

        if pre_cti_data:
            y = self.extract_region_from(array=self.dataset.pre_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"CI Pre CTI {title_str}",
                    yunit="e-",
                    filename=f"pre_cti_data{suffix}",
                ),
            )

        if signal_to_noise_map:
            y = self.extract_region_from(
                array=self.dataset.signal_to_noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Signal To Noise Map {title_str}",
                    yunit="",
                    filename=f"signal_to_noise_map{suffix}",
                ),
            )

    def subplot(
        self,
        data: bool = False,
        data_logy: bool = False,
        noise_map=False,
        signal_to_noise_map=False,
        pre_cti_data=False,
        auto_filename="subplot_dataset",
        **kwargs,
    ):
        """
        Plots the individual attributes of the plotter's `Dataset1D` object in 1D on a subplot.

        The API is such that every plottable attribute of the `Imaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is included on the subplot.

        Parameters
        ----------
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars.
        data_logy
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale.
        noise_map
            Whether to include a 1D plot (via `plot`) of the noise map.
        signal_to_noise_map
            Whether to include a 1D plot (via `plot`) of the signal-to-noise map.
        pre_cti_data
            Whether to include a 1D plot (via `plot`) of the pre-cti data.
        """

        region = kwargs.get("region", None)
        suffix = f"_{region}" if region is not None else ""

        self._subplot_custom_plot(
            data=data,
            data_logy=data_logy,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            pre_cti_data=pre_cti_data,
            auto_labels=AutoLabels(filename=f"{auto_filename}{suffix}"),
            **kwargs,
        )

    def subplot_dataset(self, region: Optional[str] = None):
        """
        Standard subplot of the attributes of the plotter's `Dataset1D` object.
        """
        self.subplot(
            region=region,
            data=True,
            noise_map=True,
            signal_to_noise_map=True,
            pre_cti_data=True,
        )



from autocti.charge_injection.fit import FitImagingCI
from autocti.model.result import ResultDataset


class ResultImagingCI(ResultDataset):
    @property
    def max_log_likelihood_full_fit(self) -> FitImagingCI:
        return self.analysis.fit_via_instance_and_dataset_from(
            instance=self.instance,
            dataset=self.analysis.dataset_full,
            hyper_noise_scale=True,
        )

    @property
    def max_log_likelihood_full_fit_no_hyper_scaling(self):
        return self.analysis.fit_via_instance_and_dataset_from(
            instance=self.instance,
            dataset=self.analysis.dataset_full,
            hyper_noise_scale=False,
        )

    @property
    def noise_scaling_map_dict(self):
        fit = self.max_log_likelihood_full_fit_no_hyper_scaling

        return {
            "regions_ci": fit.chi_squared_map_of_regions_ci,
            "parallel_eper": fit.chi_squared_map_of_parallel_eper,
            "serial_eper": fit.chi_squared_map_of_serial_eper,
            "serial_overscan_no_eper": fit.chi_squared_map_of_serial_overscan_no_eper,
        }



import numpy as np
from typing import Callable, Optional

import autoarray.plot as aplt

from autoarray.plot.auto_labels import AutoLabels

from autocti.plot.abstract_plotters import Plotter
from autocti.dataset_1d.fit import FitDataset1D


class FitDataset1DPlotter(Plotter):
    def __init__(
        self,
        fit: FitDataset1D,
        mat_plot_1d: aplt.MatPlot1D = aplt.MatPlot1D(),
        visuals_1d: aplt.Visuals1D = aplt.Visuals1D(),
        include_1d: aplt.Include1D = aplt.Include1D(),
    ):
        """
        Plots the attributes of `FitDataset1D` objects using the matplotlib method `line()` and many other matplotlib
        functions which customize the plot's appearance.

        The `mat_plot_1d` attribute wraps matplotlib function calls to make the figure. By default, the settings
        passed to every matplotlib function called are those specified in the `config/visualize/mat_wrap/*.ini` files,
        but a user can manually input values into `MatPlot1d` to customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals1D` object. Attributes may be extracted from
        the `Imaging` and plotted via the visuals object, if the corresponding entry is `True` in the `Include1D`
        object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        fit
            The fit to the dataset of a 1D dataset the plotter plots.
        mat_plot_1d
            Contains objects which wrap the matplotlib function calls that make 1D plots.
        visuals_1d
            Contains 1D visuals that can be overlaid on 1D plots.
        include_1d
            Specifies which attributes of the `ImagingCI` are extracted and plotted as visuals for 1D plots.
        """
        self.fit = fit

        super().__init__(
            dataset=fit.dataset,
            mat_plot_1d=mat_plot_1d,
            include_1d=include_1d,
            visuals_1d=visuals_1d,
        )

    def get_visuals_1d(self) -> aplt.Visuals1D:
        return self.visuals_1d

    @property
    def extract_region_from(self) -> Callable:
        return self.fit.dataset.layout.extract_region_from

    def figures_1d(
        self,
        region: Optional[str] = None,
        data: bool = False,
        data_logy: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        post_cti_data: bool = False,
        residual_map: bool = False,
        residual_map_logy: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `FitDataset1D` object in 1D.

        The API is such that every plottable attribute of the `FitDataset1D` object is an input parameter of type bool
        of the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        region
            The region on the 1D dataset where data is extracted and binned {fpr", "eper"}
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars.
        data_logy
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale.
        noise_map
            Whether to make a 1D plot (via `plot`) of the noise map.
        signal_to_noise_map
            Whether to make a 1D plot (via `plot`) of the signal-to-noise map.
        pre_cti_data
            Whether to make a 1D plot (via `plot`) of the pre-cti data.
        post_cti_data
            Whether to make a 1D plot (via `plot`) of the post-cti data.
        residual_map
            Whether to make a 1D plot (via `plot`) of the residual map, with the noise-map values included as error
            bars.
        residual_map_logy
            Whether to make a 1D plot (via `plot`) of the residual map, with the noise-map values included as error
            bars and the y-axis on a log10 scale.
        normalized_residual_map
            Whether to make a 1D plot (via `plot`) of the normalized residual map.
        chi_squared_map
            Whether to make a 1D plot (via `plot`) of the chi-squared map.
        """

        suffix = f"_{region}" if region is not None else ""
        title_str = self.title_str_from(region=region)

        y_errors = self.extract_region_from(array=self.fit.noise_map, region=region)
        y_extra = self.extract_region_from(array=self.fit.model_data, region=region)

        should_plot_zero = self.should_plot_zero_from(region=region)

        if data:
            y = self.extract_region_from(array=self.fit.data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                y_extra=y_extra,
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Data {title_str}",
                    yunit="e-",
                    filename=f"data{suffix}",
                ),
            )

        if data_logy:
            y = self.extract_region_from(array=self.fit.data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                y_extra=y_extra,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Data {title_str} [log10]",
                    yunit="e-",
                    filename=f"data_logy{suffix}",
                ),
            )

        if noise_map:
            y = self.extract_region_from(array=self.fit.noise_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="Noise-Map",
                    yunit="e-",
                    filename=f"noise_map{suffix}",
                ),
            )

        if signal_to_noise_map:
            y = self.extract_region_from(
                array=self.fit.signal_to_noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="Signal-To-Noise Map",
                    yunit="",
                    filename=f"signal_to_noise_map{suffix}",
                ),
            )

        if residual_map:
            y = self.extract_region_from(array=self.fit.residual_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Residual Map {title_str}",
                    yunit="e-",
                    filename=f"residual_map{suffix}",
                ),
            )

        if residual_map_logy:
            y = self.extract_region_from(array=self.fit.residual_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                y_extra=1.0001 * np.zeros(shape=y.shape),
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title=f"Residual Map {title_str}",
                    ylabel="e-",
                    filename=f"residual_map_logy{suffix}",
                ),
            )

        if normalized_residual_map:
            y = self.extract_region_from(
                array=self.fit.normalized_residual_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="Normalized Residual Map",
                    yunit=r"\sigma",
                    filename=f"normalized_residual_map{suffix}",
                ),
            )

        if chi_squared_map:
            y = self.extract_region_from(array=self.fit.chi_squared_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="Chi-Squared Map",
                    yunit=r"\chi^2",
                    filename=f"chi_squared_map{suffix}",
                ),
            )

        if pre_cti_data:
            y = self.extract_region_from(array=self.fit.pre_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="Pre CTI Data",
                    yunit="e-",
                    filename=f"pre_cti_data{suffix}",
                ),
            )

        if post_cti_data:
            y = self.extract_region_from(array=self.fit.post_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.get_visuals_1d(),
                auto_labels=AutoLabels(
                    title="CI Post CTI Image",
                    yunit="e-",
                    filename=f"post_cti_data{suffix}",
                ),
            )

    def subplot(
        self,
        data: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        post_cti_data: bool = False,
        residual_map: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
        auto_filename="subplot_fit",
        **kwargs,
    ):
        """
        Plots the individual attributes of the plotter's `FitDataset1D` object in 1D on a subplot.

        The API is such that every plottable attribute of the `FitDataset1D` object is an input parameter of type bool
        of the function, which if switched to `True` means that it is included on the subplot.

        Parameters
        ----------
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars.
        noise_map
            Whether to include a 1D plot (via `plot`) of the noise map.
        signal_to_noise_map
            Whether to include a 1D plot (via `plot`) of the signal-to-noise map.
        pre_cti_data
            Whether to include a 1D plot (via `plot`) of the pre-cti data.
        post_cti_data
            Whether to include a 1D plot (via `plot`) of the post-cti data.
        residual_map
            Whether to include a 1D plot (via `plot`) of the residual map.
        normalized_residual_map
            Whether to include a 1D plot (via `plot`) of the normalized residual map.
        chi_squared_map
            Whether to include a 1D plot (via `plot`) of the chi-squared map.
        """

        region = kwargs.get("region", None)
        suffix = f"_{region}" if region is not None else ""

        self._subplot_custom_plot(
            data=data,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            pre_cti_data=pre_cti_data,
            post_cti_data=post_cti_data,
            residual_map=residual_map,
            normalized_residual_map=normalized_residual_map,
            chi_squared_map=chi_squared_map,
            auto_labels=AutoLabels(
                yunit="e-",
                xlabel="Pixel No.",
                filename=f"{auto_filename}{suffix}",
            ),
        )

    def subplot_fit(self, region: Optional[str] = None):
        """
        Standard subplot of the attributes of the plotter's `FitDataset1D` object.
        """
        return self.subplot(
            region=region,
            data=True,
            signal_to_noise_map=True,
            pre_cti_data=True,
            post_cti_data=True,
            normalized_residual_map=True,
            chi_squared_map=True,
        )



from typing import Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.serial.abstract import Extract2DSerial
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DSerialEPER(Extract2DSerial):
    def region_list_from(self, settings: SettingsExtract):
        """
         Returns a list of the 2D serial EPER regions from the `region_list` containing signal  (e.g. the charge
         injection regions of charge injection data), between two input `pixels` indexes.

         Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the serial EPERs
         (e.g. the FPRs) are extracted.

         The diagram below illustrates the extraction for `pixels=(0, 1)`:

         [] = read-out electronics
         [==========] = read-out register
         [..........] = serial prescan
         [pppppppppp] = parallel overscan
         [ssssssssss] = serial overscan
         [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
         [tttttttttt] = parallel / serial charge injection region trail

                [ppppppppppppppppppppp]
                [ppppppppppppppppppppp]
            [...][ttttttttttttttttttttt][sss]
            [...][c1c1cc1c1cc1cc1ccc1cc][st1]
         |  [...][1c1c1cc1c1cc1ccc1cc1c][ts0]    |
         |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
         |  [...][0ccc0cccc0cccc0cccc0c][st1]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][ts0]   \/

         []     [=====================]
                <---------Ser--------

         The extracted regions correspond to the first serial EPER all charge injection regions:

         region_list[0] = [0, 2, 22, 225 (serial prescan is 3 pixels)
         region_list[1] = [4, 6, 22, 25] (serial prescan is 3 pixels)

         For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
         serial EPER of each charge injection region:

         array_2d_list[0] = [st0]
         array_2d_list[1] = [st1]

         Parameters
         ----------
        settings
           The settings used to extract the serial region EPERs from, which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.shape_2d[1] - self.region_list[0].x1 - settings.pixels_from_end,
                self.shape_2d[1] - self.region_list[0].x1,
            )
        return [
            region.serial_trailing_region_from(pixels=pixels)
            for region in self.region_list
        ]

    def array_2d_from(self, array: aa.Array2D) -> aa.Array2D:
        """
        Extract an arrays of all of the serial EPERs in the serial overscan region, that are to the side of a
        charge-injection scans from a charge injection array.

        The diagram below illustrates the arrays that is extracted from a array:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
        \/[...][ccccccccccccccccccccc][sts]    \/

        []     [=====================]
               <--------Ser---------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][000000000000000000000][tst]
        | [000][000000000000000000000][sts]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][000000000000000000000][tst]    | clocking
          [000][000000000000000000000][sts]    |

        []     [=====================]
               <--------Ser---------
        """

        array_2d = array.native.copy() * 0.0

        return self.add_to_array(
            new_array=array_2d,
            array=array,
            settings=SettingsExtract(pixels=(0, self.serial_overscan.total_columns)),
        )

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        """
        The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
        CTI modeling in 1D.

        This is performed by binning up the data via the `binned_array_1d_from` function.

        In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
        charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

        The function returns the this region if the 1D dataset is extracted from theserial EPERs. The
        charge region is only included if there are negative entries in the `pixels` tuple, meaning that pixels
        before the EPERs (e.g. the FPR) are extracted.

        Parameters
        ----------
        settings
           The settings used to extract the serial region EPERs from, which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """
        return extract_2d_util.binned_region_1d_eper_from(pixels=settings.pixels)



from typing import List, Optional

from autoconf import conf
from autoconf.dictable import to_dict

import autofit as af
from autoconf.dictable import output_to_json

from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D
from autocti.dataset_1d.fit import FitDataset1D
from autocti.dataset_1d.model.visualizer import VisualizerDataset1D
from autocti.dataset_1d.model.result import ResultDataset1D
from autocti.model.settings import SettingsCTI1D
from autocti.clocker.one_d import Clocker1D


class AnalysisDataset1D(af.Analysis):
    def __init__(
        self,
        dataset: Dataset1D,
        clocker: Clocker1D,
        settings_cti: SettingsCTI1D = SettingsCTI1D(),
        dataset_full: Optional[Dataset1D] = None,
    ):
        """
        Fits a CTI model to a 1D CTI dataset via a non-linear search.

        The `Analysis` class defines the `log_likelihood_function` which fits the model to the dataset and returns the
        log likelihood value defining how well the model fitted the data.

        It handles many other tasks, such as visualization, outputting results to hard-disk and storing results in
        a format that can be loaded after the model-fit is complete.

        This class is used for model-fits which fit a CTI model via a `CTI1D` object to a charge injection
        imaging dataset.

        Parameters
        ----------
        dataset
            The 1D CTI dataset that the model is fitted to.
        clocker
            The CTI arctic clocker used by the non-linear search and model-fit.
        settings_cti
            The settings controlling aspects of the CTI model in this model-fit.
        dataset_full
            The full dataset, which is visualized separate from the `dataset` that is fitted, which for example may
            not have the FPR masked and thus enable visualization of the FPR.
        """
        super().__init__()

        self.dataset = dataset
        self.clocker = clocker
        self.settings_cti = settings_cti
        self.dataset_full = dataset_full

    def region_list_from(self) -> List:
        return ["fpr", "eper"]

    def modify_before_fit(self, paths: af.DirectoryPaths, model: af.Collection):
        """
        This function is called immediately before the non-linear search begins and performs final tasks and checks
        before it begins.

        This function:

         1) Visualizes the 1D dataset, which does not change during the analysis and thus can be done once.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.
        model
            The PyAutoFit model object, which includes model components representing the galaxies that are fitted to
            the imaging data.
        """

        if paths.is_complete:
            return self

        return self

    def log_likelihood_function(self, instance: af.ModelInstance) -> float:
        """
        Determine the fitness of a particular model

        Parameters
        ----------
        instance

        Returns
        -------
        fit: Fit
            How fit the model is and the model
        """

        self.settings_cti.check_total_density_within_range(traps=instance.cti.trap_list)

        fit = self.fit_via_instance_from(instance=instance)

        return fit.log_likelihood

    def fit_via_instance_and_dataset_from(
        self, instance: af.ModelInstance, dataset: Dataset1D
    ) -> FitDataset1D:
        post_cti_data = self.clocker.add_cti(
            data=dataset.pre_cti_data, cti=instance.cti
        )

        return FitDataset1D(dataset=dataset, post_cti_data=post_cti_data)

    def fit_via_instance_from(self, instance: af.ModelInstance) -> FitDataset1D:
        return self.fit_via_instance_and_dataset_from(
            instance=instance, dataset=self.dataset
        )

    def save_attributes(self, paths: af.DirectoryPaths):
        """
        Before the model-fit via the non-linear search begins, this routine saves attributes of the `Analysis` object
        to the `pickles` folder such that they can be loaded after the analysis using PyAutoFit's database and
        aggregator tools.

        For this analysis the following are output:

        - The 1D dataset (data / noise-map / pre cti data / layout / settings etc.).
        - The mask applied to the dataset.
        - The clocker used for modeling / clocking CTI.
        - The settings used for modeling / clocking CTI.
        - The full 1D dataset (e.g. unmasked, used for visualizariton).

        It is common for these attributes to be loaded by many of the template aggregator functions given in the
        `aggregator` modules. For example, when using the database tools to reperform a fit, this will by default
        load the dataset, settings and other attributes necessary to perform a fit using the attributes output by
        this function.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization,and the pickled objects used by the aggregator output by this function.
        """

        def output_dataset(dataset, prefix):
            paths.save_fits(
                name="data",
                hdu=dataset.data.hdu_for_output,
                prefix=prefix,
            )
            paths.save_fits(
                name="noise_map",
                hdu=dataset.noise_map.hdu_for_output,
                prefix=prefix,
            )
            paths.save_fits(
                name="pre_cti_data",
                hdu=dataset.pre_cti_data.hdu_for_output,
                prefix=prefix,
            )
            paths.save_fits(
                name="mask",
                hdu=dataset.mask.hdu_for_output,
                prefix=prefix,
            )
            paths.save_json(
                name="layout",
                object_dict=to_dict(dataset.layout),
                prefix=prefix,
            )

        output_dataset(dataset=self.dataset, prefix="dataset")

        if self.dataset_full is not None:
            output_dataset(dataset=self.dataset_full, prefix="dataset_full")

        paths.save_json(
            name="clocker",
            object_dict=to_dict(self.clocker),
        )

        paths.save_json(
            name="settings_cti",
            object_dict=to_dict(self.settings_cti),
        )

    def in_ascending_fpr_order_from(self, quantity_list, fpr_value_list):
        if not conf.instance["visualize"]["general"]["general"][
            "subplot_ascending_fpr"
        ]:
            return quantity_list

        indexes = sorted(range(len(fpr_value_list)), key=lambda k: fpr_value_list[k])

        return [quantity_list[i] for i in indexes]

    def visualize_before_fit(self, paths: af.DirectoryPaths, model: af.Collection):
        region_list = self.region_list_from()

        visualizer = VisualizerDataset1D(visualize_path=paths.image_path)
        visualizer.visualize_dataset(dataset=self.dataset)
        visualizer.visualize_dataset_regions(
            dataset=self.dataset, region_list=region_list
        )

        if self.dataset_full is not None:
            visualizer.visualize_dataset(
                dataset=self.dataset_full, folder_suffix="_full"
            )
            visualizer.visualize_dataset_regions(
                dataset=self.dataset_full,
                region_list=region_list,
                folder_suffix="_full",
            )

    def visualize_before_fit_combined(
        self, analyses, paths: af.DirectoryPaths, model: af.Collection
    ):
        if analyses is None:
            return

        visualizer = VisualizerDataset1D(visualize_path=paths.image_path)

        region_list = self.region_list_from()

        dataset_list = [analysis.dataset for analysis in analyses]
        fpr_value_list = [dataset.fpr_value for dataset in dataset_list]

        dataset_list = self.in_ascending_fpr_order_from(
            quantity_list=dataset_list,
            fpr_value_list=fpr_value_list,
        )

        visualizer.visualize_dataset_combined(
            dataset_list=dataset_list,
        )
        visualizer.visualize_dataset_regions_combined(
            dataset_list=dataset_list,
            region_list=region_list,
        )

        if self.dataset_full is not None:
            dataset_full_list = [analysis.dataset_full for analysis in analyses]

            dataset_full_list = self.in_ascending_fpr_order_from(
                quantity_list=dataset_full_list,
                fpr_value_list=fpr_value_list,
            )

            visualizer.visualize_dataset_combined(
                dataset_list=dataset_full_list, folder_suffix="_full"
            )
            visualizer.visualize_dataset_regions_combined(
                dataset_list=dataset_full_list,
                region_list=region_list,
                folder_suffix="_full",
            )

    def visualize(
        self,
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis: bool,
    ):
        region_list = self.region_list_from()

        visualizer = VisualizerDataset1D(visualize_path=paths.image_path)

        fit = self.fit_via_instance_from(instance=instance)
        visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)
        visualizer.visualize_fit_regions(
            fit=fit, region_list=region_list, during_analysis=during_analysis
        )

        if self.dataset_full is not None:
            fit = self.fit_via_instance_and_dataset_from(
                instance=instance, dataset=self.dataset_full
            )
            visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)
            visualizer.visualize_fit_regions(
                fit=fit, region_list=region_list, during_analysis=during_analysis
            )

    def visualize_combined(
        self,
        analyses: List["AnalysisDataset1D"],
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis: bool,
    ):
        if analyses is None:
            return

        fit_list = [
            analysis.fit_via_instance_from(instance=instance) for analysis in analyses
        ]

        fpr_value_list = [fit.dataset.fpr_value for fit in fit_list]

        fit_list = self.in_ascending_fpr_order_from(
            quantity_list=fit_list,
            fpr_value_list=fpr_value_list,
        )

        region_list = self.region_list_from()

        visualizer = VisualizerDataset1D(visualize_path=paths.image_path)
        visualizer.visualize_fit_combined(
            fit_list=fit_list, during_analysis=during_analysis
        )
        visualizer.visualize_fit_region_combined(
            fit_list=fit_list,
            region_list=region_list,
            during_analysis=during_analysis,
        )

        if self.dataset_full is not None:
            fit_full_list = [
                analysis.fit_via_instance_and_dataset_from(
                    instance=instance, dataset=analysis.dataset_full
                )
                for analysis in analyses
            ]

            fit_full_list = self.in_ascending_fpr_order_from(
                quantity_list=fit_full_list,
                fpr_value_list=fpr_value_list,
            )

            visualizer.visualize_fit_combined(
                fit_list=fit_full_list, during_analysis=during_analysis
            )
            visualizer.visualize_fit_region_combined(
                fit_list=fit_full_list,
                region_list=region_list,
                during_analysis=during_analysis,
            )

    def make_result(
        self,
        samples: af.SamplesPDF,
    ) -> ResultDataset1D:
        return ResultDataset1D(samples=samples, analysis=self)



from astropy.io import fits
import logging
import numpy as np
import os
import shutil

from autoarray.structures.arrays.uniform_2d import Array2D

from autoarray import exc
from autoarray.structures.arrays import array_2d_util
from autoarray.layout import layout_util

from autocti.instruments.acs import acs_util

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel("INFO")


class Array2DACS(Array2D):
    """
    An ACS array consists of four quadrants ('A', 'B', 'C', 'D') which have the following layout (which are described
    at the following STScI
    link https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418).

       <--------S-----------   ---------S----------->
    [] [========= 2 =========] [========= 3 =========] []          /\
    /    [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  /        |
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | Direction arctic
    P   [xxxxxxxxx B/C xxxxxxx] [xxxxxxxxx A/D xxxxxxx]  P         | clocks an image
    |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | without any rotation
    \/  [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  \/        | (e.g. towards row 0
                                                                   | of the NumPy arrays)

    For a ACS .fits file:

    - The images contained in hdu 1 correspond to quadrants B (left) and A (right).
    - The images contained in hdu 4 correspond to quadrants C (left) and D (right).
    """

    @classmethod
    def from_fits(cls, file_path, quadrant_letter):
        """
        Use the input .fits file and quadrant letter to extract the quadrant from the full CCD, perform
        the rotations required to give correct arctic clocking and convert the image from units of COUNTS / CPS to
        ELECTRONS.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """

        hdu = acs_util.fits_hdu_via_quadrant_letter_from(
            quadrant_letter=quadrant_letter
        )

        array = array_2d_util.numpy_array_2d_via_fits_from(file_path=file_path, hdu=hdu)

        return cls.from_ccd(array_electrons=array, quadrant_letter=quadrant_letter)

    @classmethod
    def from_ccd(
        cls,
        array_electrons,
        quadrant_letter,
        header=None,
        bias_subtract_via_prescan=False,
        bias=None,
    ):
        """
        Using an input array of both quadrants in electrons, use the quadrant letter to extract the quadrant from the
        full CCD and perform the rotations required to give correct arctic.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """
        if quadrant_letter == "A":
            array_electrons = array_electrons[0:2068, 0:2072]
            roe_corner = (1, 0)
            use_flipud = True

            if bias is not None:
                bias = bias[0:2068, 0:2072]

        elif quadrant_letter == "B":
            array_electrons = array_electrons[0:2068, 2072:4144]
            roe_corner = (1, 1)
            use_flipud = True

            if bias is not None:
                bias = bias[0:2068, 2072:4144]

        elif quadrant_letter == "C":
            array_electrons = array_electrons[0:2068, 0:2072]

            roe_corner = (1, 0)
            use_flipud = False

            if bias is not None:
                bias = bias[0:2068, 0:2072]

        elif quadrant_letter == "D":
            array_electrons = array_electrons[0:2068, 2072:4144]

            roe_corner = (1, 1)
            use_flipud = False

            if bias is not None:
                bias = bias[0:2068, 2072:4144]

        else:
            raise exc.ArrayException(
                "Quadrant letter for FrameACS must be A, B, C or D."
            )

        return cls.quadrant_a(
            array_electrons=array_electrons,
            header=header,
            roe_corner=roe_corner,
            use_flipud=use_flipud,
            bias_subtract_via_prescan=bias_subtract_via_prescan,
            bias=bias,
        )

    @classmethod
    def quadrant_a(
        cls,
        array_electrons,
        roe_corner,
        use_flipud,
        header=None,
        bias_subtract_via_prescan=False,
        bias=None,
    ):
        """
        Use an input array of the left quadrant in electrons and perform the rotations required to give correct
        arctic clocking.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=roe_corner
        )

        if use_flipud:
            array_electrons = np.flipud(array_electrons)

        if bias_subtract_via_prescan:
            bias_serial_prescan_value = acs_util.prescan_fitted_bias_column(
                array_electrons[:, 18:24]
            )

            array_electrons -= bias_serial_prescan_value

            header.bias_serial_prescan_column = bias_serial_prescan_value

        if bias is not None:
            bias = layout_util.rotate_array_via_roe_corner_from(
                array=bias, roe_corner=roe_corner
            )

            if use_flipud:
                bias = np.flipud(bias)

            array_electrons -= bias

            header.bias = Array2DACS.no_mask(values=bias, pixel_scales=0.05)

        return cls.no_mask(values=array_electrons, header=header, pixel_scales=0.05)

    @classmethod
    def quadrant_b(
        cls, array_electrons, header=None, bias_subtract_via_prescan=False, bias=None
    ):
        """
        Use an input array of the right quadrant in electrons and perform the rotations required to give correct
        arctic clocking.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(1, 1)
        )

        array_electrons = np.flipud(array_electrons)

        if bias_subtract_via_prescan:
            bias_serial_prescan_value = acs_util.prescan_fitted_bias_column(
                array_electrons[:, 18:24]
            )

            array_electrons -= bias_serial_prescan_value

            header.bias_serial_prescan_column = bias_serial_prescan_value

        if bias is not None:
            bias = layout_util.rotate_array_via_roe_corner_from(
                array=bias, roe_corner=(1, 1)
            )

            bias = np.flipud(bias)

            array_electrons -= bias

            header.bias = Array2DACS.no_mask(values=bias, pixel_scales=0.05)

        return cls.no_mask(values=array_electrons, header=header, pixel_scales=0.05)

    @classmethod
    def quadrant_c(
        cls, array_electrons, header=None, bias_subtract_via_prescan=False, bias=None
    ):
        """
        Use an input array of the left quadrant in electrons and perform the rotations required to give correct
        arctic clocking.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(1, 0)
        )

        if bias_subtract_via_prescan:
            bias_serial_prescan_value = acs_util.prescan_fitted_bias_column(
                array_electrons[:, 18:24]
            )

            array_electrons -= bias_serial_prescan_value

            header.bias_serial_prescan_column = bias_serial_prescan_value

        if bias is not None:
            bias = layout_util.rotate_array_via_roe_corner_from(
                array=bias, roe_corner=(1, 0)
            )

            array_electrons -= bias

            header.bias = Array2DACS.no_mask(values=bias, pixel_scales=0.05)

        return cls.no_mask(values=array_electrons, header=header, pixel_scales=0.05)

    @classmethod
    def quadrant_d(
        cls, array_electrons, header=None, bias_subtract_via_prescan=False, bias=None
    ):
        """
        Use an input array of the right quadrant in electrons and perform the rotations required to give correct
        arctic clocking.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418
        """

        array_electrons = layout_util.rotate_array_via_roe_corner_from(
            array=array_electrons, roe_corner=(1, 1)
        )

        if bias_subtract_via_prescan:
            bias_serial_prescan_value = acs_util.prescan_fitted_bias_column(
                array_electrons[:, 18:24]
            )

            array_electrons -= bias_serial_prescan_value

            header.bias_serial_prescan_column = bias_serial_prescan_value

        if bias is not None:
            bias = layout_util.rotate_array_via_roe_corner_from(
                array=bias, roe_corner=(1, 1)
            )

            array_electrons -= bias

            header.bias = Array2DACS.no_mask(values=bias, pixel_scales=0.05)

        return cls.no_mask(values=array_electrons, header=header, pixel_scales=0.05)

    def update_fits(self, original_file_path, new_file_path):
        """
        Output the array to a .fits file.

        Parameters
        ----------
        file_path
            The path the file is output to, including the filename and the ``.fits`` extension,
            e.g. '/path/to/filename.fits'
        """

        new_file_dir = os.path.split(new_file_path)[0]

        if not os.path.exists(new_file_dir):
            os.makedirs(new_file_dir)

        if not os.path.exists(new_file_path):
            shutil.copy(original_file_path, new_file_path)

        hdulist = fits.open(new_file_path)

        hdulist[self.header.hdu].data = self.layout_2d.original_orientation_from(
            array=self
        )

        ext_header = hdulist[4].header
        bscale = ext_header["BSCALE"]

        os.remove(new_file_path)

        hdulist.writeto(new_file_path)



from typing import List

import autoarray as aa

from autocti.extract.two_d.parallel.abstract import Extract2DParallel
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DParallelFPR(Extract2DParallel):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D parallel FPR regions from the `region_list` containing signal  (e.g. the charge
        injection regions of charge injection data), between two input `pixels` indexes.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the parallel FPRs
        are extracted.

        A 2D region is defined following the convention:

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        For parallel FPR's the charge spans all columns of the charge injection region, thus the coordinates x0 and x1
        do not change. y0 and y1 are updated based on the `pixels` input.

        Negative pixel values can be input into the `pixels` tuple, whereby rows in front of the parallel FPRs are
        also extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sss]
           [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        |  [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the first parallel FPR all charge injection regions:

        region_list[0] = [0, 1, 3, 21] (serial prescan is 3 pixels)
        region_list[1] = [3, 4, 3, 21] (serial prescan is 3 pixels)

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        parallel FPR of each charge injection region:

        array_2d_list[0] = [c0c0c0cc0c0c0c0c0c0c0]
        array_2d_list[1] = [1c1c1c1c1c1c1c1c1c1c1]

        Parameters
        ----------
        settings
           The settings used to extract the parallel FPRs, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """
        return [
            region.parallel_front_region_from(
                pixels=settings.pixels, pixels_from_end=settings.pixels_from_end
            )
            for region in self.region_list
        ]

    def binned_region_1d_from(self, settings: SettingsExtract) -> aa.Region1D:
        """
        The `Extract` objects allow one to extract a `Dataset1D` from a 2D CTI dataset, which is used to perform
        CTI modeling in 1D.

        This is performed by binning up the data via the `binned_array_1d_from` function.

        In order to create the 1D dataset a `Layout1D` is required, which requires the `region_list` containing the
        charge regions on the 1D dataset (e.g. where the FPR appears in 1D after binning).

        The function returns the this region if the 1D dataset is extracted from the parallel FPRs. This is
        the full range of the `pixels` tuple, unless negative entries are included, meaning that pixels
        before the FPRs are extracted.

        Parameters
        ----------
        settings
           The settings used to extract the parallel FPR, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """
        return extract_2d_util.binned_region_1d_fpr_from(pixels=settings.pixels)
